//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					    (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalF90cxxGlue.cc	--- the Fortran90/C++ interface of the CAL
//
//	DESCRIPTION:
//		This file acts as the interface between the CAL API written in
//		Fortran90 and its implementation in C++. All platform-specifics
//		of the C++/F90 interface are centralized in this file.
//
//	CHANGE HISTORY:
//		UL, 1997-07-25: 1.0.0: - skeleton
//		UL, 1998-??-??:
//		UL, 1998-10-19: 1.2.0: - changes to version of 'rawxy2mm',
//								 'gainCorrection', 'CTIcorrection' which
//								 accept Dal column pointers (the Dal now
//								 passes pointers to Handle<Column>)
//		UL, 1998-10-20: 1.3.0: - added CTI correction for RGS
//		UL, 1998-11-02: 1.4.0: - fixed bug in CAL_getRedistribution()
//							   - minor modification of CAL_getEventPatterns
//		UL, 1999-01-14: 2.0.0: - bring in line with latest changes in src
//		UL, 1999-03-10: 2.0.1: - minor change to openCcf
//		UL, 1999-04-01: 2.1.0: - new methods: CAL_setState1/2/3 for
//								 defining state from explicit set of
//								 values/name of data set/reference to Block
//		UL, 1999-04-01: 2.2.0: - minor bug fix in modeIdToString
//		UL, 1999-05-12: 2.2.1: - changed fatal -> error and changed all
//								 error/warning tags
//		UL, 1999-06-05: 2.3.0: - many changes following changes in f90 API
//		UL, 1999-06-08: 2.3.1: - fix to make checkDope work on Dec Alpha
//							   - added CAL_setRandom
//		UL, 1999-06-20: 2.3.2: - make use of new GainCorrector/CtiCorrector
//								 classes
//		UL, 1999-07-09: 2.3.3: - added state variables `ocb' and
//								 `temperatureCamera'
//							   - added CAL_rgsOffsetCorrect()
//      UL, 1999-08-13: 2.4.0: - added new coordinate conversion routines
//                             - added CAL_piTokeV
//		UL, 1999-08-19: 2.5.0: - more validity checks in CAL_setState
//							   - dropped obsolete state variables `status'
//								 and `verbosityLevel'
//		UL, 1999-09-03: 2.6.0: - changed SER->GRADE in rgsOffsetCorrect
//		UL, 1999-09-07: 2.6.1: - added rgsGetEventThresholds
//							   - added hasMiscellaneousData
//							   - changed I/F of rgsCtiCorrect (new arguments)
//							   - added CAL_rgsGetRedistribution
//							   - changed piToKev -> piToeV
//		UL, 1999-09-16: 2.6.2: - pnReEmissionThreshold expects energies
//								 now in ADU (type changed to double)
//		UL, 1999-09-17: 2.6.3: - changed arguments of rgsGetCrossPSF + proper
//								 implementation by making use of corresponding
//								 C++ implementation
//		UL, 1999-10-11: 2.6.4: - fixed SPR 853 : non-descriptive
//								 NotImplementedYet warnings
//		UL, 1999-11-08: 2.6.5: - added CAL_getCamCoord2ToSacCoord
//		UL, 1999-11-11: 2.6.6: - bug fix in CAL_pixCoord1ToChipCoord
//		UL, 1999-11-23: 2.7.0: - added CAL_correctTime I/Fs
//							   - modified CAL_getCrossPsf I/F
//		UL, 1999-12-16: 2.8.0: - added state variable `temperatureTracking'
//		UL, 2000-01-05: 2.8.1: - named unnamed namespace
//		UL, 2000-01-13: 2.5.1: - modified CAL_omDistortion/CAL_omGetColorTrans
//		UL, 2000-02-14: 2.6.0: - bug fix in rgsGetRedistribution
//		UL, 2000-02-23: 2.7.0: - modified CAL_getFilterTransmission +
//								 CAL_getQuantumEfficiency
//							   - added CAL_getBoresightAngles/euler321toMatrix
//		UL, 2000-03-09: 2.8.0: - added CAL_getFOVmap
//		UL, 2000-03-24: 2.8.0: - dropped time arguments of
//								 CAL_pn{Gain,Cti}Correct
//		UL, 2000-04-13: 2.9.0: - proper implementation of effective area
//								 retrieval
//		UL, 2000-05-05: 2.10.0:- proper implementation of
//									CAL_getVignettingFactor
//		UL, 2000-05-12: 2.11.0:- added CAL_mosGainCorrect()
//							   - re-implemented CAL_pnReEmissionThreshold
//		UL, 2000-05-30: 2.12.0:- added alternative form of
//								 CAL_getRedistribution()
//							   - added CAL_getBinnedPIebounds()
//		UL, 2000-07-27: 2.13.0:- added CAL_omLEDtemplateMap()
//		UL, 2000-07-28: 3.6.0: - added CAL_getMOSoffsets()
//									   CAL_getMOSdarkFrameMap()
//		UL, 2000-09-19: 3.7.0: - added various PSF-related routines
//		UL, 2001-02-12: 3.9.0: - added various routines related to
//								 OM color transformation (in line with
//								 CAL HB 2.0)
//							   - added parameter empiricalLinearityCorrection
//								 to CAL_omPhotoNatural()
//							   - added CAL_omGetApertureRadius()
//		UL, 2001-03-09: 3.10.0:- added various new Boresight related stuff
//		UL, 2001-04-09: 3.11.0:- argument FRAME of CAL_pnCtiCorrect is now
//								 of type int32 (was int16)
//							   - added new state variable 'auxiliaryParameter'
//		UL, 2001-06-01: 3.12.0:- added new routines:
//									o CAL_omGetPSF()
//									o CAL_omPsfEncircledEnergy()
//									o CAL_omPsfCircleRadius()
//		UL, 2001-12-05: 3.13.0:- added new routines:
//									o CAL_getBadPixelTableCode()
//									o CAL_getAduconvCode()
//		UL, 2002-01-08: 3.14.0:- downgraded warning about subwindow'ing of
//								 BadPixels not being implemented to noisy
//								 message
//		UL, 2002-03-05: 3.15.0:- new CAL_toDirectionCosineMatrix
//								 (intended to replace CAL_getBoresightMatrix)
//		UL, 2002-06-10: 3.16.0:- new CAL_getMiscellaneousDataCcd
//		UL, 2002-07-25: 3.17.0:- new CAL_pnAdditionalGainCtiCorrect
//									 CAL_pnGainTimingCorrect
//									 CAL_pnGainBurstCorrect
//		UL, 2002-08-16: 3.18.0:- various functions related to retrieval of
//								 pattern fraction data from CCF (e.g.
//								 CAL_getPatternFractionInEnergy)
//		UL, 2002-11-18: 3.18.1:- fixed problem in CAL_pnAdditionalGainCorrect()
//								 [parOut was defined with incorrect size; lead
//								  to crash on Tru64]
//		UL, 2002-12-03: 3.19.0:- 'observationStartTime' -> 'exposureStartTime'
//							   - new argument 'exposureDuration' in
//								 CAL_setState/getState()
//		UL/RDS, 2003-10-21: 3.20.0: - added CAL_omGrismFlux/Wavelength()
//      RDS, 2003-11-26: 3.21.0: - added CAL_pnMedianMap, _pnOffsetCorrValues, _pnNoiseMap
//      RDS, 2004-08-04: 3.22.0: - added CAL_pnGainEffCorrect, CAL_pnGainFfCorrect
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Indexing of the XMM Current Calibration File, U. Lammers,
//			 15/07/97
//		[3]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/f90/CalF90cxxGlue.cc,v 1.117 2005/01/19 11:30:34 rsaxton Exp $
//
#include <cmath>
#include <cstdlib>
#include <string>

#include "common.h"
#include "CalServer.h"
#include "Celestial.h"
#include "CalF90cxxGlue.h"
#include "Ccf.h"
#include "Image.h"
#include "CalOalUtility.h"
#include "errstr.h"
#include "f90cxx.h"
#include "Dal.h"
#include "Handle.h"
#include "STime.h"

#include "BadPixelDataServer.h"
#include "CrossPsf.h"
#include "CcdQuantumDataServer.h"
#include "CtiCorrector.h"
#include "EffectiveAreaDataServer.h"
#include "EnergyCorrector.h"
#include "GeometryDataServer.h"
#include "MiscDataServer.h"
#include "OmGeoLinearizer.h"
#include "OmGrismDataServer.h"
#include "PsfDataServer.h"
#include "Redistribution.h"
#include "TimeCorrector.h"

#include <vector>

using namespace CalOalUtility;
using namespace std;

namespace CalF90CxxGlue {

bool isValidPointer(const void *p)
{
	return p && p != reinterpret_cast<void *>(0xffffffff);
}

bool checkDope(const Dope1 *dp)
{
	if (!dp)
		errstr<<"vector not properly passed"
			  <<Sas::fatal("IllegalPointerValue");

	//
	//	the following ugly cast is necessary to detect uninitialized pointers
	//	from f90 - they get set to 0xffffffff (32 bit even on Dec with
	//	64 bit pointers!)
	//
	char *uninitp = reinterpret_cast<char *>(0xffffffff);

	return dp->addr != 0 && dp->addr != uninitp;
} // checkDope()

template<class type>
inline int8 *int8ptr(type *a)	{ return reinterpret_cast<int8 *>(a); }

template<class ElementType>
ElementType *allocVec(Dope1 *dopeVec, unsigned N)
{
	ElementType *p;
	if (!checkDope(dopeVec)) {
		//	no space for data allocated yet - do it and set dope vector
		p = new ElementType[N];
		//	setup the dope stuff properly
		vector<unsigned>	oneDim(1); oneDim[0] = N;
		setupDope(*dopeVec, p, oneDim);
	} else
		p = PTR_CAST(dopeVec->addr, ElementType);

	return p;
}

}	// namespace

using namespace CalF90CxxGlue;

//
//	------------------------( CAL_openCCF )--------------------------------
//
void CXXFUNC(cal_openccf, CAL_openCCF)(const char *cifName, Integer3 len)
{
	string cifFileName = f90string(cifName, len);

	if (calServer.ccf()) {
		errstr<<"Closing previously opened CIF and opening new one `"
			  <<f90string(cifName, len)<<endl
			  <<Sas::warning("ReopeningOfCcf");
	}

	calServer.setCcf(cifFileName);

	SAS_DEBUG(clog<<"Read new CIF:\n"<<calServer.ccf()<<endl;)
} // CAL_openCCF()


//
//	------------------------( CAL_setState )-------------------------------
//
void CXXFUNC(cal_setstate1, CAL_setState1)
							(const Integer3 *instrument,
							 const Integer3 *ccdChipId,
							 const Integer3 *ccdOutNode,
							 const Integer3 *modeId,
							 const Integer3 *filterId,
							 const Integer3 *accuracyLevel,
							 const Double *temperature,
							 const Double *temperatureCamera,
							 const Double *exposureStartTime,
							 const Double *exposureDuration,
							 const Integer3 *ocb,
							 const Logical *randomize,
							 const Logical *temperatureTracking,
							 const Double *auxiliaryParameter)
{
	if (instrument)
		calServer.setInstrument(static_cast<Xmm::Instrument>(*instrument));

	CalState *state = calServer.state();
	//cout<<*state<<endl;
	#define SETVAR(var, argvp, cast)								\
		if (argvp)													\
			if (state->var()->isActive())							\
				state->var()->set(cast(*argvp));					\
			else													\
				errstr<<"State variable `"<<state->var()->name()	\
					  <<"' is not active - not altered"				\
					  <<Sas::warning("VariableDisabled");

	SETVAR(ccd, ccdChipId,);
	SETVAR(node, ccdOutNode, Xmm::toNode);
	SETVAR(mode, modeId, Xmm::toMode);
	SETVAR(filter, filterId, Xmm::toFilter);
	SETVAR(temperature, temperature,);
	SETVAR(temperatureCamera, temperatureCamera,);
	if (exposureStartTime) {
		STime t(*exposureStartTime, STime::MJD);
		state->exposureStartTime()->set(t);
	}
	SETVAR(exposureDuration, exposureDuration,);
	SETVAR(ocb, ocb,);
	SETVAR(accuracyLevel, accuracyLevel, static_cast<AccuracyLevel>);
	SETVAR(randomize, randomize,);
	SETVAR(temperatureTracking, temperatureTracking,);
	SETVAR(auxiliaryParameter, auxiliaryParameter,);

	#undef SETVAR
	//cout<<*state<<endl;
} // CAL_setState()


void CXXFUNC(cal_setstate2, CAL_setState2)
						(char *dataSetName, Integer3 length)
{
	if (!isValidPointer(dataSetName))
		errstr<<"No data set name given to CAL_setState"
			  <<Sas::error("InvalidInputArgument");

	calServer<<f90string(dataSetName, length);
} // CAL_setState()


void CXXFUNC(cal_setstate3, CAL_setState3)(Handle<DataSet> *setP)
{
	DataSet *set = setP->addressOfObject();
	if (!isValidPointer(set))
		errstr<<"No valid data set pointer given to CAL_setState"
			  <<Sas::error("InvalidInputArgument");

	calServer<<*set;
} // CAL_setState()


void CXXFUNC(cal_setstate4, CAL_setState4)(Handle<Block> *blockP)
{
	Block *block = blockP->addressOfObject();
	if (!isValidPointer(block))
		errstr<<"No valid block pointer given to CAL_setState"
			  <<Sas::error("InvalidInputArgument");

	calServer<<*block;
} // CAL_setState()



//
//	--------------------( CAL_getState )--------------------------------
//
void CXXFUNC(cal_getstate, CAL_getState)
							(Integer3 *instr,
							 Integer3 *ccdChipId,
							 Integer3 *ccdOutNode,
							 Integer3 *modeId,
							 Integer3 *filterId,
							 Integer3 *accuracyLevel,
							 Double *temp,
							 Double *tempCam,
							 Double *exposureStartTime,
							 Double *exposureDuration,
							 Integer3 *ocb,
							 Logical *randomize,
							 Logical *temperatureTracking,
							 Double *auxParam)
{
	CalState *state = calServer.state();
	if (instr)
		*instr = calServer.instrument()->id();
	if (ccdChipId)
		*ccdChipId = state->ccd()->get();
	if (ccdOutNode)
		*ccdOutNode = state->node()->get();
	if (modeId)
		*modeId = state->mode()->get();
	if (filterId)
		*filterId = state->filter()->get();
	if (temp)
		*temp = state->temperature()->get();
	if (tempCam)
		*tempCam = state->temperatureCamera()->get();
	if (exposureStartTime) {
		STime t = state->exposureStartTime()->get();
		*exposureStartTime = t.modifiedJulianDay().full();
	}
	if (exposureDuration)
		*exposureDuration = state->exposureDuration()->get();
	if (ocb)
		*ocb = state->ocb()->get();

	if (accuracyLevel)
		*accuracyLevel = state->accuracyLevel()->get();

	if (randomize)
		*randomize = state->randomize()->get();

	if (temperatureTracking)
		*temperatureTracking = state->temperatureTracking()->get();

	if (auxParam)
		*auxParam = state->auxiliaryParameter()->get();
} // CAL_getState()


//
//	------------------( CAL_setCalibrationFile )-------------------------
//
void CXXFUNC(cal_setcalibrationfile, CAL_setCalibrationFile)
							(const char *calibrationFileName,
							 Integer3 length)
{
	calServer.ccf()->registerNewCifEntry(
									f90string(calibrationFileName, length));

	SAS_DEBUG(clog<<"CIF after replacing constituent:\n"<<calServer.ccf()
			  <<endl;)
} // CAL_setCalibrationFile()


//
//	----------------------( CAL_ccfInfo )---------------------------------
//
void CXXFUNC(cal_ccfinfo, CAL_ccfInfo)
							(const char *calType,
							 struct CalInfo *calInfo,
							 Integer3 calTypeLength)
{
	CifEntry *e = calServer.ccf()->entry(calServer.instrument(),
						Ccf::toTypeId(f90string(calType, calTypeLength)),
						false);

	if (!e)
		errstr<<"The CCF does not contain a file of type: "
			  <<Xmm::toString(calServer.instrument()->id())<<"/"
			  <<f90string(calType, calTypeLength)
			  <<Sas::error("Mismatch");

	assert(calInfo);
	stringf90(Xmm::toString(e->instrument()), calInfo->instrument,
				sizeof(calInfo->instrument));
	stringf90(Ccf::toString(e->calDataType()), calInfo->calType,
				sizeof(calInfo->calType));
	stringf90(e->fileName(), calInfo->fileName, sizeof(calInfo->fileName));
	stringf90(e->submissionDate().toString(), calInfo->date,
				sizeof(calInfo->date));
	stringf90(e->validityDate().toString(), calInfo->valDate,
				sizeof(calInfo->valDate));
	stringf90(e->extensionKey(), calInfo->extSequ, sizeof(calInfo->extSequ));
	stringf90(e->extensionSequence(), calInfo->extSequId,
				sizeof(calInfo->extSequId));
	calInfo->issue = e->issue();
} // CAL_ccfInfo()


//
//	----------------( CAL_integerToReal/CAL_realToIntegr )--------------------
//
void CXXFUNC(cal_integertoreal, CAL_integerToReal)
							(const Dope1 *intVec,
							 Dope1 *realVec)
{
	checkDope(intVec);
	unsigned vecSize = intVec->dim[0].extent;

	real32 *rp = allocVec<real32>(realVec, vecSize);
	CalInt16Vector iv(calServer.int16Vector(SHORTP(intVec->addr), vecSize));
	CalReal32Vector rv(calServer.real32Vector(rp, vecSize));
	iv.toReal(rv);
} // CAL_integerToReal()


void CXXFUNC(cal_realtointeger, CAL_realToInteger)
							(const Dope1 *realVec,
							 Dope1 *intVec)
{
	checkDope(realVec);
	unsigned vecSize = realVec->dim[0].extent;

	int16 *ip = allocVec<int16>(intVec, vecSize);
	CalReal32Vector rv(calServer.real32Vector(FLOATP(realVec->addr), vecSize));
	CalInt16Vector iv(calServer.int16Vector(ip, vecSize));
	rv.toInt(iv);
} // CAL_realToInteger()


//
//	-------------------------( CAL_setRandom )------------------------------
//
void CXXFUNC(cal_setrandom, CAL_setRandom)
							(const Double *min, const Double *max)
{
	assert(min && max);
	calServer.setRandom(*min, *max);
} // CAL_setRandom()




#define RELDATA(a) if (a) { delete[] a->addr; a->addr = 0; }
#define RELEASEMEM(fname, cname, N)		\
	void CXXFUNC(fname, cname)			\
		(Dope##N *a1, Dope##N *a2, Dope##N *a3) \
		{ RELDATA(a1); RELDATA(a2); RELDATA(a3); }

RELEASEMEM(cal_releasememoryint3onedim, CAL_releaseMemoryInt3OneDim, 1)
RELEASEMEM(cal_releasememorylogicalonedim, CAL_releaseMemoryLogicalOneDim, 1)
RELEASEMEM(cal_releasememoryint3twodim, CAL_releaseMemoryInt3TwoDim, 2)
RELEASEMEM(cal_releasememoryreal2onedim, CAL_releaseMemoryReal2OneDim, 1)
RELEASEMEM(cal_releasememoryreal2twodim, CAL_releaseMemoryReal2TwoDim, 2)

RELEASEMEM(cal_releasememoryint1onedim, CAL_releaseMemoryInt1OneDim, 1)
RELEASEMEM(cal_releasememoryint2onedim, CAL_releaseMemoryInt2OneDim, 1)
RELEASEMEM(cal_releasememoryreal1onedim, CAL_releaseMemoryReal1OneDim, 1)
RELEASEMEM(cal_releasememoryint2twodim, CAL_releaseMemoryInt2TwoDim, 2)
RELEASEMEM(cal_releasememoryreal1twodim, CAL_releaseMemoryReal1TwoDim, 2)

RELEASEMEM(cal_releasememoryint1threedim, CAL_releaseMemoryInt1ThreeDim, 3)

#undef RELEASEMEM
#undef RELDATA

//
//	-----------------(CAL_getBadPixelTableCode/getAduconvCode )-------------
//
void CXXFUNC(cal_getbadpixeltablecode, CAL_getBadPixelTableCode)
							(char *codeStr, Integer3 length)
{
	BadPixelDataServer *bpds = 0; bpds = calServer.getAtom(bpds);
	stringf90(bpds->badPixelTableCode(), codeStr, length);
} // CAL_getBadPixelTableCode()


void CXXFUNC(cal_getaduconvcode, CAL_getAduconvCode)
							(char *codeStr, Integer3 length)
{
	EnergyCorrector *aduConv = 0; aduConv = calServer.getAtom(aduConv);
	stringf90(aduConv->badPixelTableCode(), codeStr, length);
} // CAL_getAduconvCode()


//
//	------------------( CAL_getBadPixelList )--------------------------
//
void CXXFUNC(cal_getbadpixellist, CAL_getBadPixelList)
							(Dope1 *rawX, Dope1 *rawY, Dope1 *type,
							 Dope1 *yextent, Dope1 *uplinked,
 							 const Integer2 *rawX0,
							 const Integer2 *rawY0,
							 const Integer2 *rawXsize,
							 const Integer2 *rawYsize)
{
	if (rawX0 || rawY0 || rawXsize || rawYsize)
		errstr<<"Support for subwindows in extraction of bad pixel lists "
				"not implemented yet"
			  <<Sas::message(Msg::Lib, Msg::Noisy);

	BadPixelDataServer *bpds = 0; bpds = calServer.getAtom(bpds);
	BadPixelMap map = bpds->map();
	unsigned nPix = map.size();

	//	allocate space for vectors
	Integer2 *rawxList		= new Integer2[nPix];
	Integer2 *rawyList		= new Integer2[nPix];
	Integer2 *yextentList	= new Integer2[nPix];
	Integer2 *typeList		= new Integer2[nPix];
	Logical  *uplinkedList	= new Logical[nPix];

	//	setup the dope stuff properly
	vector<unsigned>	oneDim(1); oneDim[0] = nPix;
	setupDope(*rawX, rawxList, oneDim);
	setupDope(*rawY, rawyList, oneDim);
	setupDope(*yextent, yextentList, oneDim);
	setupDope(*type, typeList, oneDim);
	setupDope(*uplinked, uplinkedList, oneDim);

	if (!nPix)
		return;
	BadPixelMap::const_iterator i;
	unsigned n = 0;
	for (i=map.begin(); i!=map.end(); ++i) {
		rawxList[n]			= static_cast<Integer2>((*i).first.first);
		rawyList[n]			= static_cast<Integer2>((*i).first.second);
		yextentList[n]		= static_cast<Integer2>((*i).second.yextent());
		typeList[n]			= static_cast<Integer2>((*i).second.type());
		if (calServer.instrument()->id() == Xmm::OM) {
			// for OM we encode also the severity in the badpix type
			const BadPixelOm &bp =
							static_cast<const BadPixelOm &>((*i).second);
			typeList[n] += 10 * static_cast<Integer2>(bp.severity());
		}

		uplinkedList[n++]	= static_cast<Logical>((*i).second.uplinked());
	}
} // CAL_getBadPixelList()


//
//	---------------------( CAL_isPixelBad )--------------------------------
//
Logical CXXFUNC(cal_ispixelbad, CAL_isPixelBad)
							(const Integer2 *rawX, const Integer2 *rawY,
							 Integer2 *yextent, Integer2 *defect,
							 Logical *uplinked)
{
	BadPixelDataServer *bpds = 0; bpds = calServer.getAtom(bpds);
	BadPixelMap pixmap = bpds->map();
	const BadPixel *pix = pixmap.isPixelBad(PixelLocation(*rawX, *rawY));
	if (pix) {
		if (yextent)
			*yextent = static_cast<Integer2>(pix->yextent());
		if (defect)
			*defect = static_cast<Integer2>(pix->type());
		if (uplinked)
			*uplinked = static_cast<Logical>(pix->uplinked());

		return static_cast<Logical>(true);
	}

	return static_cast<Logical>(false);
} // CAL_isPixelBad()


//
//	---------------------( CAL_getBadPixelMap )---------------------------
//
void CXXFUNC(cal_getbadpixelmap, CAL_getBadPixelMap)(Dope2 *badPixMap,
					 const Integer2 *rawX0, const Integer2 *rawY0,
					 const Integer2 *rawXsize, const Integer2 *rawYsize)
{
	if (rawX0 || rawY0 || rawXsize || rawYsize);
		errstr<<"Support for subwindows in extraction of bad pixel map "
				"not yet implemented"
			  <<Sas::message(Msg::Lib, Msg::Noisy);

	unsigned extX = calServer.instrument()->maxX() -
					calServer.instrument()->minX() + 1,
			 extY = calServer.instrument()->maxY() -
					calServer.instrument()->minY() + 1;

	I16Image ccdMap = I16Image(extX, extY, SHORT(0));

	BadPixelDataServer *bpds = 0; bpds = calServer.getAtom(bpds);
	BadPixelMap map = bpds->map();
	unsigned nPix = map.size();
	if (nPix)
		for (BadPixelMap::const_iterator i=map.begin();
		 i!=map.end(); ++i) {
			unsigned x = (*i).first.first,
					 y = (*i).first.second;

			short	 val = (*i).second.type();
			if ((*i).second.uplinked())
				val += 10;

			for (unsigned n=0; n<(*i).second.yextent(); ++n)
				if (x <= extX && y+n <= extY)
					ccdMap[x-1][y+n-1] = val;
				else
					errstr<<"illegal pixel address ("<<x<<"/"<<y<<")"
						  <<Sas::warning("IllegalValues");
		}

	//
	//	setup the dope stuff
	//
	short *data = new Integer2[ccdMap.size()];
	memcpy(data, ccdMap.data(), ccdMap.memSize());
	vector<unsigned> twoDim(2); twoDim[0] = extX; twoDim[1] = extY;
	setupDope(*badPixMap, data, twoDim);
} // CAL_getBadPixelMap()


//
//	---( CAL_getBoresight / CAL_withSacCoordOrientation /
//		 CAL_projectOntoTelCoord )-----------------------------------------
//
namespace CalF90CxxGlueBoresight {
	static double bstime = 0.;
}

Handle<Boresight> CXXFUNC(cal_getboresight, CAL_getBoresight)
							 (const Double *t)
{
	CalF90CxxGlueBoresight::bstime = *t;
	Boresight *bs = 0; bs = calServer.getAtom(bs);
	return Handle<Boresight>(bs);
} // CAL_getBoresight()


void CXXFUNC(cal_withsaccoordorientation, CAL_withSacCoordOrientation)
							(const Handle<Boresight> *boresight,
							 const SpacecraftAttitudeType *att)
{
	assert(boresight);
	assert(att);

	boresight->addressOfObject()->withSacCoordOrientation(
		SpacecraftAttitude(att->ra, att->dec, att->apos));
} // CAL_withSacCoordOrientation()


void CXXFUNC(cal_projectontotelcoord, CAL_projectOntoTelCoord)
							(const Handle<Boresight> *boresight,
							 const Dope1 *srcDirs,
							 Dope1 *theta,
							 Dope1 *phi)
{
	assert(boresight);
	checkDope(srcDirs);
	unsigned N = srcDirs->dim[0].extent;
	EquatorialDirectionType *srcs = (EquatorialDirectionType *)srcDirs->addr;
	vector<EquatorialDirection> srcVec;
	for (unsigned n=0; n<N; ++n)
		srcVec.push_back(EquatorialDirection(RightAscension(srcs[n].ra),
											  Declination(srcs[n].dec)));

	//	do the projection
	XmmTelescopePointList tpl =
			boresight->addressOfObject()->projectOntoTelCoord(srcVec);
	//	copy results to output vector

	real64 *thetaData = allocVec<real64>(theta, N),
		   *phiData = allocVec<real64>(phi, N);

	memcpy(thetaData, tpl.theta().ptr(), N*sizeof(real64));
	memcpy(phiData, tpl.phi().ptr(), N*sizeof(real64));
} // CAL_projectOntoTelCoord()


//
//	---------------( CAL_getBoresightMatrix/Angles )-----------------------
//
void CXXFUNC(cal_getboresightmatrix, CAL_getBoresightMatrix)
						(const Double *time, Dope2 *matrix)
{
	Boresight *bs = 0; bs = calServer.getAtom(bs);
	AttitudeMatrix a = bs->matrix(*time);
	Double *data = new Double[a.size()];
	memcpy(data, a.dataFlat(), a.memSize());
	vector<unsigned> twoDim(2); twoDim[0] = 3, twoDim[1] = 3;
	setupDope(*matrix, data, twoDim);
} // CAL_getBoresightMatrix()


Dope2 CXXFUNC(cal_todirectioncosinematrix1, CAL_toDirectionCosineMatrix1)
							(const Handle<Boresight> *boresight)
{
	assert(boresight);
	AttitudeMatrix a =
		boresight->addressOfObject()->matrix(CalF90CxxGlueBoresight::bstime);
	Double *data = new Double[a.size()];
	memcpy(data, a.dataFlat(), a.memSize());
	vector<unsigned> twoDim(2); twoDim[0] = 3, twoDim[1] = 3;
	Dope2 matrix;
	setupDope(matrix, data, twoDim);

	return matrix;
} // CAL_toDirectionCosineMatrix()


Dope2 CXXFUNC(cal_todirectioncosinematrix2, CAL_toDirectionCosineMatrix2)
							(const BSangles *eulerAngles)
{
	assert(eulerAngles);

	EulerAngles ea(eulerAngles->phi, eulerAngles->theta, eulerAngles->psi,
				   EulerAngles::Euler321);
	AttitudeMatrix a = ea.toMatrix();
	Double *data = new Double[a.size()];
	memcpy(data, a.dataFlat(), a.memSize());
	vector<unsigned> twoDim(2); twoDim[0] = 3, twoDim[1] = 3;
	Dope2 matrix;
	setupDope(matrix, data, twoDim);

	return matrix;
} // CAL_toDirectionCosineMatrix()


extern "C" BSangles CXXFUNC(cal_toeulerangles, CAL_toEulerAngles)
							(const Handle<Boresight> *boresight)
{
	assert(boresight);
	BSangles bsa;
	const EulerAngles &ea =
		boresight->addressOfObject()->angles(CalF90CxxGlueBoresight::bstime);
	bsa.phi = ea.phi();
	bsa.theta = ea.theta();
	bsa.psi = ea.psi();

	return bsa;
} // CAL_toEulerAngles()


BSangles CXXFUNC(cal_getboresightangles, CAL_getBoresightAngles)
						(const Double *time)
{
	assert(time);

	Boresight *bs = 0; bs = calServer.getAtom(bs);
	BSangles bsa;
	const EulerAngles &ea = bs->angles(*time);
	bsa.phi = ea.phi();
	bsa.theta = ea.theta();
	bsa.psi = ea.psi();

	return bsa;
} // CAL_getBoresightMatrix()


//
//	---------------( CAL_euler321toMatrix )----------------------------------
//
void CXXFUNC(cal_euler321tomatrix, CAL_euler321toMatrix)
							(const BSangles *eulerAngles, Double *matrix)
{
	if (!eulerAngles || !matrix)
        errstr<<"Argument `eulerAnges' or `matrix' is missing - cannot "
				"continue"<<Sas::error("MissingInputArguments");

	EulerAngles ea(eulerAngles->phi, eulerAngles->theta, eulerAngles->psi,
				   EulerAngles::Euler321);
	AttitudeMatrix mat = ea.toMatrix();
	const double *psrc = mat.dataFlat();
	for (unsigned n=0; n<9; ++n)
		matrix[n] = psrc[n];
} // CAL_euler321toMatrix()


//
//	----------------------( CAL_getCamCoord2ToSacCoord )----------------------
//
void CXXFUNC(cal_getcamcoord2tosaccoord, CAL_getCamCoord2ToSacCoord)
							(const Double *time, Dope2 *matrix)
{
	assert(time);

	GeometryDataServer *cf = 0; cf = calServer.getAtom(cf);
	AttitudeMatrix a = cf->coordFrameQuantities().cc2ToSacCoordMatrix;
	Double *data = new Double[a.size()];
	memcpy(data, a.dataFlat(), a.memSize());
	vector<unsigned> twoDim(2); twoDim[0] = 3, twoDim[1] = 3;
	setupDope(*matrix, data, twoDim);
} // CAL_getCamCoord2ToSacCoord()


//
//	----------------------( CAL_getDiffuseXBackground )---------------------
//
void CXXFUNC(cal_getdiffusexbackground, CAL_getDiffuseXBackground)
				(const Double *time, const Double *theta, const Double *phi,
				 Dope1 *spectrum, Dope1 *piVec)
{
	BackgroundDataServer *background = 0;
	background = calServer.getAtom(background);
	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);

	CalReal64Vector thetaVec(const_cast<Double *>(theta), 1),
					phiVec(const_cast<Double *>(phi), 1);

	double xmm, ymm, zmm;
	CalReal64Vector xmmVec(&xmm, 1), ymmVec(&ymm, 1), zmmVec(&zmm, 1);
	XmmTelescopePointList tpl = ct->telCoordList(thetaVec, phiVec);
	XmmCartesianPointList cpl = tpl.toCamCoord2(xmmVec, ymmVec, zmmVec);

	double r = sqrt(CalOalUtility::SQR(xmm) +
					CalOalUtility::SQR(ymm));

	BackgroundSpectrum diffuseBackSpec(background->diffuseX(*time, r));
	const vector<unsigned> &channels = diffuseBackSpec.channels();

	unsigned nChan = diffuseBackSpec.size();
	Single *spec		= new Single[nChan];
	Integer2 *channs	= new Integer2[nChan];

	for (unsigned n=0; n<nChan; ++n) {
		spec[n] = diffuseBackSpec[n];
		channs[n] = channels[n];
	}

	//	setup the dope stuff properly
	vector<unsigned>	oneDim(1); oneDim[0] = nChan;
	setupDope(*spectrum, spec, oneDim);
	setupDope(*piVec, channs, oneDim);
} // CAL_getDiffuseXBackground()



//
//	----------------------( CAL_getParticleBackground )---------------------
//
void CXXFUNC(cal_getparticlebackground, CAL_getParticleBackground)
				(const Double *time, const Double *theta, const Double *phi,
				 Dope1 *spectrum, Dope1 *piVec)
{
	BackgroundDataServer *background = 0;
	background = calServer.getAtom(background);
	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);

	CalReal64Vector thetaVec(const_cast<Double *>(theta), 1),
					phiVec(const_cast<Double *>(phi), 1);

	double xmm, ymm, zmm;
	CalReal64Vector xmmVec(&xmm, 1), ymmVec(&ymm, 1), zmmVec(&zmm, 1);
	XmmTelescopePointList tpl = ct->telCoordList(thetaVec, phiVec);
	XmmCartesianPointList cpl = tpl.toCamCoord2(xmmVec, ymmVec, zmmVec);

	double r = sqrt(CalOalUtility::SQR(xmm) +
					CalOalUtility::SQR(ymm));

	BackgroundSpectrum particleBackSpec(background->particle(*time, r));
	const vector<unsigned> &channels = particleBackSpec.channels();

	unsigned nChan = particleBackSpec.size();
	Single *spec		= new Single[nChan];
	Integer2 *channs	= new Integer2[nChan];

	for (unsigned n=0; n<nChan; ++n) {
		spec[n] = particleBackSpec[n];
		channs[n] = channels[n];
	}

	//	setup the dope stuff properly
	vector<unsigned>	oneDim(1); oneDim[0] = nChan;
	setupDope(*spectrum, spec, oneDim);
	setupDope(*piVec, channs, oneDim);
} // CAL_getParticleBackground()



//
//	----------------------( CAL_has/getMiscellaneousData )---------------------
//
Logical CXXFUNC(cal_hasmiscellaneousdata, CAL_hasMiscellaneousData)
					(const char *keyword, Integer3 length)
{
	MiscDataServer *miscData = 0; miscData = calServer.getAtom(miscData);
	return static_cast<Logical>(miscData->has(string(keyword, length)));
} // CAL_hasMiscellaneousData()

Logical CXXFUNC(cal_hasmiscellaneousdataccd, CAL_hasMiscellaneousDataCcd)
					(const char *keyword, Integer3 length)
{
	MiscDataServer *miscData = 0; miscData = calServer.getAtom(miscData);
	return static_cast<Logical>(
						miscData->hasWithCcdNumber(string(keyword, length)));
} // CAL_hasMiscellaneousDataCcd()

void CXXFUNC(cal_getmiscellaneousdata, CAL_getMiscellaneousData)
					(const char *keyword, Double *value, Integer3 length)
{
	MiscDataServer *miscData = 0; miscData = calServer.getAtom(miscData);
	string s(keyword, length);
	*value = miscData->get(s);
} // CAL_getMiscellaneousData()


Double CXXFUNC(cal_getmiscellaneousdatavalue, CAL_getMiscellaneousDataValue)
					(const char *keyword, Integer3 length)
{
	Double value;
	CXXFUNC(cal_getmiscellaneousdata, CAL_getMiscellaneousData)(
			keyword, &value, length);

	return value;		   
} // CAL_getMiscellaneousDataValue()

Double CXXFUNC(cal_getmiscellaneousdataccd, CAL_getMiscellaneousDataCcd)
					(const char *keywordStem, Integer3 length)
{
	MiscDataServer *miscData = 0; miscData = calServer.getAtom(miscData);
	string s(keywordStem, length);
	return miscData->getWithCcdNumber(s);
} // CAL_getMiscellaneousDataCcd()


//
//	-----------------( CAL_getContaminationData )--------------------------
//
void CXXFUNC(cal_getcontaminationdata, CAL_getContaminationData)
					(const Double *time,
					 ContaminationData *contaminationData)
{
	assert(time);
	errstr<<"Retrieval of contamination data from CCF not implemented yet"
		  <<Sas::warning("NotImplementedYet");
	contaminationData->TBS = 0;
} // CAL_getContaminationData()

//
//	--------------( CAL_getEbounds/binnedPIebounds/eBins )---------------------
//
#define EBOUNDS(name1, name2, method)								\
void CXXFUNC(name1, name2)											\
				(Dope1 *channel, Dope1 *Emin, Dope1 *Emax)			\
{																	\
	CanonicalRedist *rmf = 0; rmf = calServer.getAtom(rmf);					\
	const BinnedAxis &ebounds = rmf->method();						\
	unsigned nChan = ebounds.bins();								\
	Integer2 *channelVec = new Integer2[nChan];						\
	Double *eminVec		 = new Double[nChan],						\
		   *emaxVec		 = new Double[nChan];						\
																	\
	for (unsigned i=0; i<nChan; ++i) {								\
		channelVec[i] = i;											\
		eminVec[i] = ebounds.bin(i).lo();							\
		emaxVec[i] = ebounds.bin(i).hi();							\
	}																\
																	\
	vector<unsigned>	oneDim(1); oneDim[0] = nChan;				\
	setupDope(*channel, channelVec, oneDim);						\
	setupDope(*Emin, eminVec, oneDim);								\
	setupDope(*Emax, emaxVec, oneDim);								\
} // CAL_getEbounds()


EBOUNDS(cal_getebounds, CAL_getEbounds, eBounds)
EBOUNDS(cal_getbinnedpiebounds, CAL_getBinnedPIebounds, binnedPIeBounds)
EBOUNDS(cal_getebins, CAL_getEbins, eBins)

#undef EBOUNDS


//
//	--------------------( CAL_getRedistribution )---------------------------
//
void CXXFUNC(cal_getredistribution1, CAL_getRedistribution1)
				(const Double *E, Dope1 *response, Integer3 *pixeltype,
				 Integer2 *peakChannel)
{
	if (pixeltype && *pixeltype < 0)
		errstr<<"Illegal value for pixeltype argument"
			  <<Sas::error("IllegalValues");

	CanonicalRedist *rmf = 0; rmf = calServer.getAtom(rmf);
	Spectrum spec = pixeltype ? rmf->spectrum(*E, EventGrades(*pixeltype))
							  : rmf->spectrum(*E);

	unsigned nChans = spec.size();
	Double *respVec = new Double [nChans];

	vector2cvec(spec, respVec, nChans);

	//	setup dope vectors
	vector<unsigned>	oneDim(1); oneDim[0] = nChans;
	setupDope(*response, respVec, oneDim);

	if (peakChannel)
		*peakChannel = static_cast<Integer2>(spec.peakChannel());
} // CAL_getRedistribution()


void CXXFUNC(cal_getredistribution2, CAL_getRedistribution2)
				(const Double *eIn,
				 const Dope1 *eAxisLoBounds,
				 const Dope1 *eAxisHiBounds,
				 Dope1 *response, Integer3 *pixeltype,
				 Integer2 *peakChannel)
{
	if (pixeltype && *pixeltype < 0)
		errstr<<"Illegal value for pixeltype argument"
			  <<Sas::error("IllegalValues");

	assert(eAxisLoBounds && eAxisHiBounds);
	unsigned Nlo = eAxisLoBounds->dim[0].extent,
			 Nhi = eAxisHiBounds->dim[0].extent;
	if (Nlo != Nhi)
		errstr<<"Vectors with lower and upper bin boundaries must have the "
				"same size"<<Sas::error("VectorSizeMismatch");

	CalReal64Vector eLower(DOUBLEP(eAxisLoBounds->addr), Nlo),
					eUpper(DOUBLEP(eAxisHiBounds->addr), Nhi);
	NonUniformilyBinnedAxis eAxis(eLower, true, eUpper[Nhi-1]);

	CanonicalRedist *rmf = 0; rmf = calServer.getAtom(rmf);
	Spectrum spec = pixeltype ? rmf->spectrum(*eIn, eAxis,
											  EventGrades(*pixeltype))
							  : rmf->spectrum(*eIn, eAxis);

	unsigned nChans = spec.size();
	Double *respVec = new Double [nChans];

	vector2cvec(spec, respVec, nChans);

	//	setup dope vectors
	vector<unsigned>	oneDim(1); oneDim[0] = nChans;
	setupDope(*response, respVec, oneDim);

	if (peakChannel)
		*peakChannel = static_cast<Integer2>(spec.peakChannel());
} // CAL_getRedistribution()


//
//	---------------------( CAL_getDarkFrame )----------------------------
//
void CXXFUNC(cal_getdarkframe, CAL_getDarkFrame)(Dope2 *map)
{
	errstr<<"not yet implemented - map set to 0"
		  <<Sas::warning("NoyYetImplemented");

	unsigned extX = calServer.instrument()->maxX() -
					calServer.instrument()->minX() + 1,
			 extY = calServer.instrument()->maxY() -
					calServer.instrument()->minY() + 1;

	Single *darkMap = new Single[extX * extY];
	for (unsigned n=0; n<extX * extY; darkMap[n++] = 0.f);

	//
	//	setup the dope stuff
	//
	vector<unsigned> twoDim(2); twoDim[0] = extX; twoDim[1] = extY;
	setupDope(*map, darkMap, twoDim);
} // CAL_getDarkFrame()


//
//	-------------------------( CAL_getBackgroundMap )------------------------
//
void CXXFUNC(cal_getbackgroundmap, CAL_getBackgroundMap)
							(const Double *time,
							 Dope2 *backMap)
{
	assert(time);
	errstr<<"Retrieval of background map from CCF not implemented yet - will "
			"return default 0-map"
		  <<Sas::warning("NotImplementedYet");

	unsigned extX = calServer.instrument()->maxX() -
					calServer.instrument()->minX() + 1,
			 extY = calServer.instrument()->maxY() -
					calServer.instrument()->minY() + 1;
	Single *backMapVec = new Single[extX * extY];
	for (unsigned n=0; n<extX * extY; backMapVec[n++] = 0.f);

	//
	//	setup the dope stuff
	//
	vector<unsigned> twoDim(2); twoDim[0] = extX; twoDim[1] = extY;
	setupDope(*backMap, backMapVec, twoDim);

	errstr<<"background map set to 0"
		  <<Sas::warning("DefaultValues");
} // CAL_getBackgroundMap()


//
//	--------------------( CAL_getFilterTransmission )------------------------
//
Double CXXFUNC(cal_getfiltertransmission1, CAL_getFilterTransmission1)
							(const Double *E,
							 const Double *theta, const Double *phi)
{
	FilterDataServer *fds = 0; fds = calServer.getAtom(fds);
	const FilterTransmissionCurve &fc = fds->at(*theta, *phi);

	return fc.transmission(*E);
} // CAL_getFilterTransmission()


Double CXXFUNC(cal_getfiltertransmission2, CAL_getFilterTransmission2)
							(const Double *Efrom, const Double *Eto,
							 const Double *theta, const Double *phi)
{
	FilterDataServer *fds = 0; fds = calServer.getAtom(fds);
	const FilterTransmissionCurve &fc = fds->at(*theta, *phi);

	return fc.transmission(*Efrom, *Eto);
} // CAL_getFilterTransmission()


//
//	---------------------( CAL_getFOVmap )--------------------------------
//
void CXXFUNC(cal_getfovmap, CAL_getFOVmap)(Dope2 *map,
					 const Integer2 *rawX0, const Integer2 *rawY0,
					 const Integer2 *rawXsize, const Integer2 *rawYsize)
{
	unsigned minX = rawX0 ? *rawX0 : calServer.instrument()->minX(),
			 minY = rawY0 ? *rawY0 : calServer.instrument()->minY(),
			 maxX = rawXsize ? minX + *rawXsize - 1
							 : calServer.instrument()->maxX(),
			 maxY = rawYsize ? minY + *rawYsize - 1
							 : calServer.instrument()->maxY(),
			 extX = maxX - minX + 1,
			 extY = maxY - minY + 1;

	unsigned N = extX * extY;
	int16 xv[N], yv[N];
	unsigned n = 0;
	for (unsigned x=minX; x<=maxX; ++x)
		for (unsigned y=minY; y<=maxY; ++y) {
			xv[n] = x;
			yv[n++] = y;
		}

	CalInt16Vector rawX(xv, N),
				   rawY(yv, N),
				   chipX,
				   chipY;
	chipX.clone(rawX);
	chipY.clone(rawY);

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
	XmmPixelList pl = ct->pixCoord1List(chipX, chipY);
	vector<bool> inFov = pl.insideFOV();

	I16Image fovMap = I16Image(extX, extY, SHORT(0));
	for (unsigned n=0; n<N; ++n)
		if (inFov[n])
			fovMap[xv[n]-minX][yv[n]-minY] = 1;

	//
	//	setup the dope stuff
	//
	short *data = new Integer2[fovMap.size()];
	memcpy(data, fovMap.data(), fovMap.memSize());
	vector<unsigned> twoDim(2); twoDim[0] = extX; twoDim[1] = extY;
	setupDope(*map, data, twoDim);
} // CAL_getFOVmap()


//
//	-----------( CAL_getMOSoffsets / CAL_getMOSdarkFrameMap )---------------
//
void CXXFUNC(cal_getmosoffsets, CAL_getMOSoffsets)
							(Dope1 *offX, Dope1 *offY)
{
	DarkFrameDataServer *dfs = 0; dfs = calServer.getAtom(dfs);
	const vector<int16> &colVec = dfs->columnOffsetVector(),
						&rowVec = dfs->rowOffsetVector();

	int16 *offXp;
	unsigned nColVec = colVec.size();
	offXp = checkDope(offX) ? SHORTP(offX->addr) : new int16[nColVec];
	vector2cvec(colVec, offXp, nColVec);
	if (!checkDope(offX)) {
		//	setup the dope stuff properly
		vector<unsigned>	oneDim(1); oneDim[0] = nColVec;
		setupDope(*offX, offXp, oneDim);
	}

	int16 *offYp;
	unsigned nRowVec = rowVec.size();
	offYp = checkDope(offY) ? SHORTP(offY->addr) : new int16[nRowVec];
	vector2cvec(rowVec, offYp, nRowVec);
	if (!checkDope(offY)) {
		//	setup the dope stuff properly
		vector<unsigned>	oneDim(1); oneDim[0] = nRowVec;
		setupDope(*offY, offYp, oneDim);
	}
} // CAL_getMOSoffsets()


void CXXFUNC(cal_getmosdarkframemap, CAL_getMOSdarkFrameMap)
							(Dope2 *map)
{
	DarkFrameDataServer *dfs = 0; dfs = calServer.getAtom(dfs);
	const SImage &dfmap = dfs->darkFrameMap();
	Single *data = new Single[dfmap.size()];
	unsigned n=0,
			 NX = dfmap.xExtent(),
			 NY = dfmap.yExtent();
	for (unsigned nx=0; nx<NX; ++nx)
		for (unsigned ny=0; ny<NY; ++ny)
			data[n++] = dfmap[ny][nx];
	vector<unsigned> twoDim(2); twoDim[0] = dfmap.yExtent(),
								twoDim[1] = dfmap.xExtent();
	setupDope(*map, data, twoDim);
} // CAL_getMOSdarkFrameMap()


//
//	-----------------------( CAL_getModeParameters )------------------------
//
void CXXFUNC(cal_getmodeparameters, CAL_getModeParameters)
				(ModeParameterData *modeParameterData)
{
	ModeParameters *modeParam = 0; modeParam = calServer.getAtom(modeParam);
	modeParameterData->integrationTime = modeParam->integrationTime();
	modeParameterData->shiftTime = modeParam->shiftTime();
	modeParameterData->frameTime = modeParam->frameTime();
	modeParameterData->cycleTime = modeParam->cycleTime();
	modeParameterData->timeResolution = modeParam->timeResolution();
	modeParameterData->dutyCycle = modeParam->dutyCycle();
	modeParameterData->smearedFraction = modeParam->smearedFraction();
	modeParameterData->pointSourcePileUp = modeParam->pointSourcePileUp();
	modeParameterData->limitingFlux = modeParam->limitingFlux();

	modeParameterData->underscanX = modeParam->underX();
	modeParameterData->overscanX = modeParam->overX();
	modeParameterData->underscanY = modeParam->underY();
	modeParameterData->overscanY = modeParam->overY();
	modeParameterData->windowX0 = modeParam->windowX0();
	modeParameterData->windowY0 = modeParam->windowY0();
	modeParameterData->windowDX = modeParam->windowDX();
	modeParameterData->windowDY = modeParam->windowDY();
} // CAL_getModeParameters()


//
//	------------------------( CAL_getVignettingFactor )--------------------
//
Double CXXFUNC(cal_getvignettingfactor, CAL_getVignettingFactor)
							(const Double *E,
							 const Double *theta,
							 const Double *phi)
{
	if (!E || !theta || !phi)
        errstr<<"Argument `E' or `theta' or `phi' is missing - cannot "
				"continue"<<Sas::error("MissingInputArguments");

	static bool first = true;
	static VignettingCurve vignc;
	static double thetaLast = 0.;
	static double phiLast = 0.;
    Xmm::Instrument instid = calServer.instrument()->id();
    static Xmm::Instrument instlast;

  /* Calculate the vignetting curve if this is the first time through
     or if the theta, phi or instrument parameters have changed */
	if (first || *theta != thetaLast || *phi != phiLast || 
           instid != instlast) {
		first = false;
		thetaLast = *theta;
		phiLast = *phi;
        instlast = calServer.instrument()->id();
		EffectiveAreaDataServer *aeds = 0; aeds = calServer.getAtom(aeds);

		vignc = aeds->vignettingCurve(*theta, *phi);
	}

	return vignc(*E);
} // CAL_getVignettingFactor()


//
//	--------------------( CAL_{rgs}getEffectiveArea )--------------------------
//
static EffectiveAreaCurve &effectiveAreaCurve(double theta, double phi)
{
	static bool first = true;
	static EffectiveAreaCurve aec;
	static double thetaLast = 0.;
	static double phiLast = 0.;

	if (first || theta != thetaLast || phi != phiLast) {
		first = false;
		thetaLast = theta;
		phiLast = phi;
		EffectiveAreaDataServer *aeds = 0; aeds = calServer.getAtom(aeds);

		aec = aeds->effectiveAreaCurve(theta, phi);
	}

	return aec;
} // effectiveAreaCurve()

Double CXXFUNC(cal_geteffectivearea1, CAL_getEffectiveArea1)
							(const Double *E,
							 const Double *theta,
							 const Double *phi)
{
	static EffectiveAreaDataServer *aeds = 0;
	if (!aeds) {
		EffectiveAreaDataServer *ds = 0; ds = calServer.getAtom(aeds);
		aeds = ds;
	}

	return aeds->effectiveArea(*E, *theta, *phi);
} // CAL_getEffectiveArea()


Double CXXFUNC(cal_geteffectivearea2, CAL_getEffectiveArea2)
							(const Double *Efrom,
							 const Double *Eto,
							 const Double *theta,
							 const Double *phi)
{
	return effectiveAreaCurve(*theta, *phi).area(*Efrom, *Eto);
} // CAL_getEffectiveArea()


static EffectiveAreaCurve &rgsEffectiveAreaCurve(int order,
												 double theta, double phi)
{
	static bool first = true;
	static EffectiveAreaCurve aec;
	static double thetaLast = 0.;
	static double phiLast = 0.;
	static int orderLast = 0;

	if (first ||
		theta != thetaLast || phi != phiLast || orderLast != order) {
		first = false;
		thetaLast = theta;
		phiLast = phi;
		orderLast = order;
		EffectiveAreaDataServer *aeds = 0; aeds = calServer.getAtom(aeds);

		aec = aeds->effectiveAreaCurve(order, theta, phi);
	}

	return aec;
} // rgsEffectiveAreaCurve()

Double CXXFUNC(cal_rgsgeteffectivearea1, CAL_rgsGetEffectiveArea1)
							(const Double *E,
							 const Integer3 *order,
							 const Double *theta,
							 const Double *phi)
{
	return rgsEffectiveAreaCurve(*order, *theta, *phi).area(*E);
} // CAL_rgsGetEffectiveArea()

Double CXXFUNC(cal_rgsgeteffectivearea2, CAL_rgsGetEffectiveArea2)
							(const Double *Efrom,
							 const Double *Eto,
							 const Double *theta,
							 const Double *phi,
							 const Integer3 *order)
{
	return rgsEffectiveAreaCurve(*order, *theta, *phi).area(*Efrom, *Eto);
} // CAL_rgsGetEffectiveArea()


//
//	----------------------( CAL_getHKwindows )-----------------------------
//
void CXXFUNC(cal_gethkwindows, CAL_getHKwindows)(Dope1 *windowList)
{
	assert(windowList);
	errstr<<"Retrieval of HK acceptance windows from CCF not yet implemented "
			"yet - will return nothing"
		  <<Sas::warning("NotImplementedYet");
} // CAL_getHKwindows()


//
//	---------------------( CAL_getEventPatterns )--------------------------
//
void CXXFUNC(cal_geteventpatterns, CAL_getEventPatterns)
		(Dope3 *patterns, Integer2 *eduThreshold)
{
	PatternLibServer *pattLib = 0; pattLib = calServer.getAtom(pattLib);

	unsigned nr = pattLib->patterns();
	const EventPattern &p0 = pattLib->pattern(0);

	Integer1 *evpatt = new Integer1[nr * p0.memSize()];
	for (unsigned n=0; n<nr; ++n)
		memcpy(evpatt + n*p0.memSize(),
			   pattLib->pattern(n).data(), p0.memSize());

	//	setup dope vectors
	vector<unsigned>	threeDim(3);
	threeDim[0] = nr;
	threeDim[1] = p0.xExtent();
	threeDim[2] = p0.yExtent();

	setupDope(*patterns, evpatt, threeDim);

	if (eduThreshold)
		*eduThreshold = 1;
} // CAL_getEventPatterns()


//
//	------------------( CAL_getPatternFraction* )------------------------------
//
Double CXXFUNC(cal_curvevalueat, CAL_curveValueAt)
						(const Handle<CcdQuantumEfficiencyCurve> *curve,
						 const Double *abscissaValue)
{
	assert(curve && abscissaValue);
	return curve->addressOfObject()->efficiency(*abscissaValue);
} // CAL_curveValueAt()


Double CXXFUNC(cal_curveweightedintegralover, CAL_curveWeightedIntegralOver)
						(const Handle<CcdQuantumEfficiencyCurve> *curve,
						 const Double *fromValue, const Double *toValue)
{
	assert(curve && fromValue && toValue);
	return curve->addressOfObject()->efficiency(*fromValue, *toValue);
} // CAL_curveWeightedIntegralOver()


Handle<PatternFractionDataServer>
	  CXXFUNC(cal_getpatternfractionserver, CAL_getPatternFractionServer)()
{
	CcdQuantumDataServer *qds = 0; qds = calServer.getAtom(qds);
	const PatternFractionDataServer *pfds =
									qds->patternFractionData().operator->();

	return Handle<PatternFractionDataServer>(pfds);
} // CAL_getPatternFractionServer()


void CXXFUNC(cal_patternfractionat, CAL_patternFractionAt)
						(const Handle<PatternFractionDataServer> *pfds,
						 const Integer2 *chipX, const Integer2 *chipY)
{
	assert(pfds && chipX && chipY);

	pfds->addressOfObject()->at(*chipX, *chipY);
} // CAL_patternFractionAt()


Handle<CcdQuantumEfficiencyCurve>
			  CXXFUNC(cal_getpatternfractiontotalqe,
					  CAL_getPatternFractionTotalQE)
						(const Handle<PatternFractionDataServer> *pfds)
{
	assert(pfds);

	return Handle<CcdQuantumEfficiencyCurve>(pfds->addressOfObject()->total());
} // CAL_getPatternFractionTotalQE()


Handle<CcdQuantumEfficiencyCurve>
			  CXXFUNC(cal_getpatternfractioninenergy,
					  CAL_getPatternFractionInEnergy)
						(const Handle<PatternFractionDataServer> *pfds,
						 const Integer3 *grade)
{
	assert(pfds && grade);

	const PatternFractionDataServer *server = pfds->addressOfObject();
	server->grade(EventGrades(*grade));

	return server->inEnergySpace();
} // CAL_getPatternFractionInEnergy()


Dope1 CXXFUNC(cal_getpatternfractioninchannel, CAL_getPatternFractionInChannel)
						(const Handle<PatternFractionDataServer> *pfds,
						 const Integer3 *grade)
{
	assert(pfds && grade);

	const PatternFractionDataServer *server = pfds->addressOfObject();
	server->grade(EventGrades(*grade));

	const vector<double> &frac = server->inChannelSpace();
	Dope1 dope;
	real64 *dp = allocVec<real64>(&dope, frac.size());
	vector2cvec(frac, dp, frac.size());

	return dope;
} // CAL_getPatternFractionInChannel()


//
//	------------------( CAL_getQuantumEfficiency )--------------------------
//
Double CXXFUNC(cal_getquantumefficiency1, CAL_getQuantumEfficiency1)
							(const Double *E,
							 const Integer2 *rawX, const Integer2 *rawY,
							 const Integer3 *patternId)
{
	CcdQuantumDataServer *qds = 0; qds = calServer.getAtom(qds);
	const CcdQuantumEfficiencyCurve &ef =
				qds->at(*rawX, *rawY).pattern(patternId ? *patternId : 0);

	return ef.efficiency(*E);
} // CAL_getQuantumEfficiency()

Double CXXFUNC(cal_getquantumefficiency2, CAL_getQuantumEfficiency2)
							(const Double *Efrom, const Double *Eto,
							 const Integer2 *rawX, const Integer2 *rawY,
							 const Integer3 *patternId)
{
	CcdQuantumDataServer *qds = 0; qds = calServer.getAtom(qds);
	const CcdQuantumEfficiencyCurve &ef =
				qds->at(*rawX, *rawY).pattern(patternId ? *patternId : 0);

	return ef.efficiency(*Efrom, *Eto);
} 

/* Return mean size of large patterns at energy E */
Double CXXFUNC(cal_getlargepatternsize1, CAL_getLargePatternSize1)
							(const Double *E)
{
	CcdQuantumDataServer *qds = 0; qds = calServer.getAtom(qds);
	const CcdLargePatternCurve &lp =
				qds->sizeofLargePatterns();

	return lp.efficiency(*E);
} 

/* Return mean size of large patterns between energies E1 and E2 */
Double CXXFUNC(cal_getlargepatternsize2, CAL_getLargePatternSize2)
			(const Double *Efrom, const Double *Eto)
{
	CcdQuantumDataServer *qds = 0; qds = calServer.getAtom(qds);
	const CcdLargePatternCurve &lp =
				qds->sizeofLargePatterns();

	return lp.efficiency(*Efrom, *Eto);
} // CAL_getLargePatterns()

//
//	--------------------( CAL_getPsf / CAL_psf* )-----------------------------
//
Handle<Psf> CXXFUNC(cal_getpsf, CAL_getPSF)
							(const Double *E,
							 const Double *theta,
							 const Double *phi)
{
	assert(E && theta && phi);
    PsfDataServer *pds = 0; pds = calServer.getAtom(pds);
    const Psf &psf = pds->psf(*E, *theta, *phi);

	return Handle<Psf>(&psf);
} // CAL_getPSF()


Double CXXFUNC(cal_psfpointvalue, CAL_psfPointValue)
							(const Handle<Psf> *psf,
							 const Double *x,
							 const Double *y)
{
	assert(psf && x && y);
	return psf->addressOfObject()->operator()(*x, *y);
} // CAL_psfPointValue()


Logical CXXFUNC(cal_psfbinsizechangeable, CAL_psfBinSizeChangeable)
							(const Handle<Psf> *psf)
{
	assert(psf);
	return psf->addressOfObject()->binSizeChangeable();
} // CAL_psfBinSizeChangeable()


void CXXFUNC(cal_psfsetbinsize, CAL_psfSetBinSize)
							(const Handle<Psf> *psf,
							 const PsfBinSize *binSize)
{
	assert(psf && binSize);

	psf->addressOfObject()->binSize(Psf::BinSize(binSize->binX,
												 binSize->binY));
} // CAL_psfSetBinSize()


PsfBinSize CXXFUNC(cal_psfbinsize, CAL_psfBinSize)
							(const Handle<Psf> *psf)
{
	assert(psf);
	PsfBinSize binSize;
	binSize.binX = psf->addressOfObject()->binSize().binX();
	binSize.binY = psf->addressOfObject()->binSize().binY();

	return binSize;
} // CAL_psfBinSize()


PsfValidityRanges CXXFUNC(cal_psfvalidityranges, CAL_psfValidityRanges)()
{
	PsfValidityRanges valRanges;
	PsfDataServer *pds = 0; pds = calServer.getAtom(pds);
	valRanges.eMin = pds->eValidity().lo();
	valRanges.eMax = pds->eValidity().hi();
	valRanges.thetaMin = pds->thetaValidity().lo();
	valRanges.thetaMax = pds->thetaValidity().hi();
	valRanges.phiMin = pds->phiValidity().lo();
	valRanges.phiMax = pds->phiValidity().hi();

	return valRanges;
} // CAL_psfValidityRanges()


void CXXFUNC(cal_psfimage, CAL_psfImage)
							(const Handle<Psf> *psf,
							 const PsfWindow *window,
							 Dope2 *map)
{
	assert(psf && window && map);

	Image<float> image = psf->addressOfObject()->
						image(Psf::PsfWindow(window->halfX, window->halfY));

	Single *mapData = new Single[image.size()];
	memcpy(mapData, image.data(), image.memSize());

	//
	//	setup the dope stuff
	//
	vector<unsigned> twoDim(2); twoDim[0] = image.xExtent(),
								twoDim[1] = image.yExtent();
	setupDope(*map, mapData, twoDim);
} // CAL_psfImage()


PsfWindow CXXFUNC(cal_psfenboxedenergy, CAL_psfEnboxedEnergy)
							(const Handle<Psf> *psf,
							 const Double *fraction,
							 const Double *aspectRatio)
{
	assert(psf && fraction);
	double ar = aspectRatio ? *aspectRatio : 1.;
	Psf::PsfWindow window = psf->addressOfObject()->enclosingBox(*fraction,
																  ar);
	PsfWindow win;
	win.halfX = window.halfX();
	win.halfY = window.halfY();

	return win;
} // CAL_psfEnboxedEnergy()


Double CXXFUNC(cal_psfencircledenergy, CAL_psfEncircledEnergy)
							(const Handle<Psf> *psf,
							 const Double *fraction)
{
	assert(psf && fraction);
	return psf->addressOfObject()->enclosingCircle(*fraction);
} // CAL_psfEncircledEnergy()


//
//	--------------------( CAL_correctTime )---------------------------------
//
void CXXFUNC(cal_correcttime1, CAL_correctTime1)
							(const Dope1 *timeCoordinate)
{
	assert(timeCoordinate);

	CalReal64Vector timeVec(DOUBLEP(timeCoordinate->addr),
							timeCoordinate->dim[0].extent);

	//
	//	get the atom and perform the correction
	//
	TimeCorrector *tc = 0; tc = calServer.getAtom(tc);
	tc->correct(timeVec);
} // CAL_correctTime()


void CXXFUNC(cal_correcttime2, CAL_correctTime2)
							(const Dope1 *timeCoordinate,
							 const Dope1 *rawY)
{
	assert(timeCoordinate);
	assert(rawY);

	unsigned N = timeCoordinate->dim[0].extent;
	CalReal64Vector timeVec(DOUBLEP(timeCoordinate->addr), N);
	CalInt16Vector rawYvec(SHORTP(rawY->addr), N);

	//
	//	get the atom and perform the correction
	//
	TimeCorrector *tc = 0; tc = calServer.getAtom(tc);
	tc->correct(timeVec, rawYvec);
} // CAL_correctTime()


void CXXFUNC(cal_correcttime3, CAL_correctTime3)
							(const Dope1 *timeCoordinate,
							 const Dope1 *rawY,
							 Integer2 *sourceY)
{
	assert(timeCoordinate);
	assert(rawY);
	assert(sourceY);

	unsigned N = timeCoordinate->dim[0].extent;
	CalReal64Vector timeVec(DOUBLEP(timeCoordinate->addr), N);
	CalInt16Vector rawYvec(SHORTP(rawY->addr), N);

	//
	//	get the atom and perform the correction
	//
	TimeCorrector *tc = 0; tc = calServer.getAtom(tc);
	tc->correct(timeVec, rawYvec, *sourceY);
} // CAL_correctTime()


//
//	-------------------------( CAL_rawxy2mm )------------------------------
//
void CXXFUNC(cal_rawxy2mm, CAL_rawXY2mm)
					(const Dope1 *rawX, const Dope1 *rawY,
					 Dope1 *Xmm, Dope1 *Ymm, Dope1 *Zmm,
					 const Integer3 *coordType)
{
	checkDope(rawX);
	checkDope(rawY);
	unsigned vecSize = rawX->dim[0].extent;
	real64 *xmm = allocVec<real64>(Xmm, vecSize),
		   *ymm = allocVec<real64>(Ymm, vecSize),
		   *zmm = allocVec<real64>(Zmm, vecSize);

	CalInt16Vector rawXVec(0, vecSize),
				   rawYVec(0, vecSize);
	int16	*xv = SHORTP(rawX->addr),
			*yv = SHORTP(rawY->addr);
	//	copy input vector because data get overwritten in PIXCOORD1->CHIPCOORD
	//	conversion
	for (unsigned n=0; n<vecSize; ++n) {
		rawXVec[n] = xv[n];
		rawYVec[n] = yv[n];
	}

	CalReal64Vector xmmVec(xmm, vecSize),
			 	    ymmVec(ymm, vecSize),
				    zmmVec(zmm, vecSize);

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
	XmmCartesianPointList cpl = ct->pixCoord1List(rawXVec, rawYVec).
			toChipCoord().toCamCoord1(xmmVec, ymmVec, zmmVec);

	unsigned toFrame = 0;
	if (coordType)
		toFrame = *coordType;
	switch(toFrame) {
		case 0:	//	to SACOORD
				cpl.toCamCoord2().toSacCoord();
				break;
		case 1: //	to CAMCOORD1
				break;
		case 2: //	to CAMCOORD2
				cpl.toCamCoord2();
				break;
		default:
			errstr<<"Conversion only to SACCOORD/CAMCOORD1/CAMCOORD2 supported"
				  <<" - "<<toFrame<<" is unknown"
				  <<Sas::error("UnsupportedOperation");
			break;
	}
} // CAL_rawXY2mm()


//
//	-----------------( CAL_pixCoord0ToPixCoord1 )------------------------------
//
void CXXFUNC(cal_pixcoord0topixcoord1, CAL_pixCoord0ToPixCoord1)
							(const Dope1 *telX,
							 const Dope1 *telY,
							 Dope1 *rawX,
							 Dope1 *rawY)
{
	checkDope(telX);
	checkDope(telY);
	unsigned vecSize = telX->dim[0].extent;

	int16 *rawXdata = allocVec<int16>(rawX, vecSize),
		  *rawYdata = allocVec<int16>(rawY, vecSize);

	CalInt16Vector telXvec(SHORTP(telX->addr), vecSize),
				   telYvec(SHORTP(telY->addr), vecSize),
				   rawXvec(rawXdata, vecSize),
				   rawYvec(rawYdata, vecSize);

	for (unsigned n=0; n<vecSize; ++n) {
		rawXvec[n] = telXvec[n];
		rawYvec[n] = telYvec[n];
	}

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
	XmmPixelList pl = ct->pixCoord0List(rawXvec, rawYvec).toPixCoord1();
} // CAL_pixCoord0ToPixCoord1()


//
//	-----------( CAL_pixCoord1ToChipCoord / CAL_chipCoordToPixCoord1 )--------
//
void CXXFUNC(cal_pixcoord1tochipcoord, CAL_pixCoord1ToChipCoord)
							(const Dope1 *rawX,
							 const Dope1 *rawY,
							 Dope1 *chipX,
							 Dope1 *chipY,
							 const Dope1 *nodes)
{
	checkDope(rawX);
	checkDope(rawY);
	unsigned vecSize = rawX->dim[0].extent;

	int16 *chipXdata = allocVec<int16>(chipX, vecSize),
		  *chipYdata = allocVec<int16>(chipY, vecSize);

	CalInt16Vector rawXvec(SHORTP(rawX->addr), vecSize),
				   rawYvec(SHORTP(rawY->addr), vecSize),
				   chipXvec(chipXdata, vecSize),
				   chipYvec(chipYdata, vecSize);

	for (unsigned n=0; n<vecSize; ++n) {
		chipXvec[n] = rawXvec[n];
		chipYvec[n] = rawYvec[n];
	}

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
    XmmPixelList pl = ct->pixCoord1List(chipXvec, chipYvec);
    if (nodes) {
        CalInt8Vector nodeVec(int8ptr(nodes->addr), vecSize);
        pl.toChipCoord(&nodeVec);
    } else
        pl.toChipCoord();
} // CAL_pixCoord1ToChipCoord()


void CXXFUNC(cal_chipcoordtopixcoord1, CAL_chipCoordToPixCoord1)
							(const Dope1 *chipX,
							 const Dope1 *chipY,
							 Dope1 *rawX,
							 Dope1 *rawY,
                             Dope1 *nodes)
{
	checkDope(chipX);
	checkDope(chipY);
	unsigned vecSize = chipX->dim[0].extent;

	int16 *rawXdata = allocVec<int16>(rawX, vecSize),
		  *rawYdata = allocVec<int16>(rawY, vecSize);

	CalInt16Vector chipXvec(SHORTP(chipX->addr), vecSize),
				   chipYvec(SHORTP(chipY->addr), vecSize),
				   rawXvec(rawXdata, vecSize),
				   rawYvec(rawYdata, vecSize);

	for (unsigned n=0; n<vecSize; ++n) {
		rawXvec[n] = chipXvec[n];
		rawYvec[n] = chipYvec[n];
	}

    int8 *nodeData = 0;
	if (nodes)
		nodeData = allocVec<int8>(nodes, vecSize);
    CalInt8Vector nodesVec(nodeData, vecSize);

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
    XmmPixelList pl = ct->chipCoordList(rawXvec, rawYvec).
                                                    toPixCoord1(nodesVec);
} // CAL_chipCoordToPixCoord1()


//
//	-----------( CAL_chipCoordToCamCoord1 / CAL_camCoord1ToChipCoord )--------
//
void CXXFUNC(cal_chipcoordtocamcoord1, CAL_chipCoordToCamCoord1)
							(const Dope1 *chipX,
							 const Dope1 *chipY,
							 Dope1 *xmm1,
							 Dope1 *ymm1,
							 Dope1 *zmm1)
{
	checkDope(chipX);
	checkDope(chipY);
	unsigned vecSize = chipX->dim[0].extent;

	real64 *xmmData = allocVec<real64>(xmm1, vecSize),
		   *ymmData = allocVec<real64>(ymm1, vecSize),
		   *zmmData = allocVec<real64>(zmm1, vecSize);

	CalInt16Vector chipXvec(SHORTP(chipX->addr), vecSize),
				   chipYvec(SHORTP(chipY->addr), vecSize);
    CalReal64Vector xmm1vec(xmmData, vecSize),
                    ymm1vec(ymmData, vecSize),
                    zmm1vec(zmmData, vecSize);

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
	XmmCartesianPointList pl = ct->chipCoordList(chipXvec, chipYvec).
                                        toCamCoord1(xmm1vec, ymm1vec, zmm1vec);
} // CAL_chipCoordToCamCoord1()


void CXXFUNC(cal_camcoord1tochipcoord, CAL_camCoord1ToChipCoord)
							(const Dope1 *xmm1,
							 const Dope1 *ymm1,
							 const Dope1 *zmm1,
							 Dope1 *chipX,
							 Dope1 *chipY,
							 Dope1 *ccds)
{
	checkDope(xmm1);
	checkDope(ymm1);
	checkDope(zmm1);
	unsigned vecSize = xmm1->dim[0].extent;

	int16 *chipXdata = allocVec<int16>(chipX, vecSize),
		  *chipYdata = allocVec<int16>(chipY, vecSize);

    CalReal64Vector xmm1vec(DOUBLEP(xmm1->addr), vecSize),
                    ymm1vec(DOUBLEP(ymm1->addr), vecSize),
                    zmm1vec(DOUBLEP(zmm1->addr), vecSize);
	CalInt16Vector chipXvec(chipXdata, vecSize),
				   chipYvec(chipYdata, vecSize);

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
	XmmCartesianPointList cpl = ct->camCoord1List(xmm1vec, ymm1vec, zmm1vec);
	XmmPixelList pl = cpl.toChipCoord(chipXvec, chipYvec);

	if (ccds) {
		int8 *ccdData = new int8 [vecSize];
		vector2cvec(cpl.onCcd(), ccdData, vecSize);

		vector<unsigned>	oneDim(1); oneDim[0] = vecSize;
		setupDope(*ccds, ccdData, oneDim);
	}
} // CAL_camCoord1ToChipCoord()


#define CART_TO_CART(nameLC, nameUP,                                    \
                     initRoutine, convRoutine)                          \
void CXXFUNC(nameLC, nameUP)(const Dope1 *xin,                          \
							 const Dope1 *yin,                          \
							 const Dope1 *zin,                          \
							 Dope1 *xout,                               \
							 Dope1 *yout,                               \
							 Dope1 *zout)                               \
{                                                                       \
	checkDope(xin);                                                     \
	checkDope(yin);                                                     \
	checkDope(zin);                                                     \
	unsigned vecSize = xin->dim[0].extent;                              \
	real64 *xoutdata = allocVec<real64>(xout, vecSize),					\
		   *youtdata = allocVec<real64>(yout, vecSize),					\
		   *zoutdata = allocVec<real64>(zout, vecSize);					\
                                                                        \
    CalReal64Vector xinvec(DOUBLEP(xin->addr), vecSize),                \
                    yinvec(DOUBLEP(yin->addr), vecSize),                \
                    zinvec(DOUBLEP(zin->addr), vecSize),                \
                    xoutvec(xoutdata, vecSize),                         \
                    youtvec(youtdata, vecSize),                         \
                    zoutvec(zoutdata, vecSize);                         \
                                                                        \
    for (unsigned n=0; n<vecSize; ++n) {                                \
        xoutvec[n] = xinvec[n];                                         \
        youtvec[n] = yinvec[n];                                         \
        zoutvec[n] = zinvec[n];                                         \
    }                                                                   \
                                                                        \
	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);             \
	XmmCartesianPointList pl = ct->initRoutine(xoutvec, youtvec, zoutvec).\
                                        convRoutine();                  \
                                                                        \
} // CAL_realCoordToRealCoord()


//
//	-----------( CAL_camCoord1ToCamCoord2 / CAL_camCoord2ToCamCoord1 )--------
//
CART_TO_CART(cal_camcoord1tocamcoord2, CAL_camCoord1ToCamCoord2,
             camCoord1List, toCamCoord2)

CART_TO_CART(cal_camcoord2tocamcoord1, CAL_camCoord2ToCamCoord1,
             camCoord2List, toCamCoord1)


//
//	-----------( CAL_camCoord2ToSacCoord / CAL_sacCoordToCamCoord2 )--------
//
CART_TO_CART(cal_camcoord2tosaccoord, CAL_camCoord2ToSacCoord,
             camCoord2List, toSacCoord)

CART_TO_CART(cal_saccoordtocamcoord2, CAL_sacCoordToCamCoord2,
             sacCoordList, toCamCoord2)


//
//	-----------( CAL_sacCoordToRowCoord / CAL_rowCoordToSacCoord )--------
//
void CXXFUNC(cal_saccoordtorowcoord, CAL_sacCoordToRowCoord)
							(const Dope1 *xmmSc,
							 const Dope1 *ymmSc,
							 const Dope1 *zmmSc,
							 Dope1 *beta,
							 Dope1 *chi)
{
	checkDope(xmmSc);
	checkDope(ymmSc);
	checkDope(zmmSc);
	unsigned vecSize = xmmSc->dim[0].extent;

	real64 *betaData = allocVec<real64>(beta, vecSize),
		   *chiData = allocVec<real64>(chi, vecSize);

    CalReal64Vector xmmScvec(DOUBLEP(xmmSc->addr), vecSize),
                    ymmScvec(DOUBLEP(ymmSc->addr), vecSize),
                    zmmScvec(DOUBLEP(zmmSc->addr), vecSize);
	CalReal64Vector betavec(betaData, vecSize),
                    chivec(chiData, vecSize);

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
	XmmRowlandPointList pl = ct->sacCoordList(xmmScvec, ymmScvec, zmmScvec).
                                        toRowCoord(betavec, chivec);
} // CAL_sacCoordToRowCoord()


void CXXFUNC(cal_rowcoordtosaccoord, CAL_rowCoordToSacCoord)
							(const Dope1 *beta,
							 const Dope1 *chi,
							 Dope1 *xmmSc,
							 Dope1 *ymmSc,
							 Dope1 *zmmSc)
{
	checkDope(beta);
	checkDope(chi);
	unsigned vecSize = beta->dim[0].extent;

	real64 *xmmScData = allocVec<real64>(xmmSc, vecSize),
		   *ymmScData = allocVec<real64>(ymmSc, vecSize),
		   *zmmScData = allocVec<real64>(zmmSc, vecSize);

    CalReal64Vector betavec(DOUBLEP(beta->addr), vecSize),
                    chivec(DOUBLEP(chi->addr), vecSize);
	CalReal64Vector xmmScvec(xmmScData, vecSize),
                    ymmScvec(ymmScData, vecSize),
                    zmmScvec(zmmScData, vecSize);

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
	XmmCartesianPointList pl = ct->rowCoordList(betavec, chivec).
                                    toSacCoord(xmmScvec, ymmScvec, zmmScvec);
} // CAL_rowCoordToSacCoord()


//
//	-----------( CAL_camCoord2ToTelCoord / CAL_telCoordToCamCoord2 )--------
//
void CXXFUNC(cal_camcoord2totelcoord, CAL_camCoord2ToTelCoord)
							(const Dope1 *xmm2,
							 const Dope1 *ymm2,
							 const Dope1 *zmm2,
							 Dope1 *theta,
							 Dope1 *phi)
{
	checkDope(xmm2);
	checkDope(ymm2);
	checkDope(zmm2);
	unsigned vecSize = xmm2->dim[0].extent;

	real64 *thetaData = allocVec<real64>(theta, vecSize),
		   *phiData = allocVec<real64>(phi, vecSize);

    CalReal64Vector xmm2vec(DOUBLEP(xmm2->addr), vecSize),
                    ymm2vec(DOUBLEP(ymm2->addr), vecSize),
                    zmm2vec(DOUBLEP(zmm2->addr), vecSize),
                    thetavec(thetaData, vecSize),
                    phivec(phiData, vecSize);

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
	XmmTelescopePointList pl = ct->camCoord2List(xmm2vec, ymm2vec, zmm2vec).
                                        toTelCoord(thetavec, phivec);
} // CAL_camCoord2ToTelCoord()


void CXXFUNC(cal_telcoordtocamcoord2, CAL_telCoordToCamCoord2)
							(const Dope1 *theta,
							 const Dope1 *phi,
							 Dope1 *xmm2,
							 Dope1 *ymm2,
							 Dope1 *zmm2)
{
	checkDope(theta);
	checkDope(phi);
	unsigned vecSize = theta->dim[0].extent;

	real64 *xmm2Data = allocVec<real64>(xmm2, vecSize),
		   *ymm2Data = allocVec<real64>(ymm2, vecSize),
		   *zmm2Data = allocVec<real64>(zmm2, vecSize);

    CalReal64Vector thetavec(DOUBLEP(theta->addr), vecSize),
                    phivec(DOUBLEP(phi->addr), vecSize),
                    xmm2vec(xmm2Data, vecSize),
                    ymm2vec(ymm2Data, vecSize),
                    zmm2vec(zmm2Data, vecSize);

	GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
	XmmCartesianPointList pl = ct->telCoordList(thetavec, phivec).
                                    toCamCoord2(xmm2vec, ymm2vec, zmm2vec);
} // CAL_telCoordToCamCoord2()


//
//	-------------------------( CAL_offsetCorrect )-----------------------------
//
void CXXFUNC(cal_offsetcorrect, CAL_offsetCorrect)
							(const Dope1 *time,
							 const Dope1 *eIn,
							 Dope1 *eOut,
							 const Dope1 *nodes)
{
	checkDope(time);
	checkDope(eIn);
	unsigned vecSize = eIn->dim[0].extent;

	real32 *rp = allocVec<real32>(eOut, vecSize);
	CalReal64Vector timeVec(DOUBLEP(time->addr), vecSize);
	CalReal32Vector	eInVec(calServer.real32Vector(FLOATP(eIn->addr),vecSize)),
					eOutVec(calServer.real32Vector(rp, vecSize));

	//
	//	get the atom
	//
	EnergyCorrector *offsetCorrector = 0;
	offsetCorrector = calServer.getAtom(offsetCorrector);

	if (nodes)
		if (checkDope(nodes)) {
			CalInt8Vector nodesVec(int8ptr(nodes->addr), vecSize);

			offsetCorrector->offsetCorrect(timeVec, nodesVec, eInVec, eOutVec);
		} else
			errstr<<"nodes vector is empty"
				  <<Sas::fatal("EmptyVector");
	else
		offsetCorrector->offsetCorrect(timeVec, eInVec, eOutVec);
} // CAL_offsetCorrect()



//
//	-----------------------( CAL_gainCorrect )-------------------------------
//
void CXXFUNC(cal_gaincorrect, CAL_gainCorrect)
							(const Dope1 *time,
							 const Dope1 *eIn,
							 Dope1 *eOut,
							 const Dope1 *nodes)
{
	checkDope(time);
	checkDope(eIn);
	unsigned vecSize = eIn->dim[0].extent;

	real32 *rp = allocVec<real32>(eOut, vecSize);
	CalReal64Vector timeVec(DOUBLEP(time->addr), vecSize);
	CalReal32Vector	eInVec(calServer.real32Vector(FLOATP(eIn->addr),vecSize)),
					eOutVec(calServer.real32Vector(rp, vecSize));

	//
	//	get the atom
	//
	EnergyCorrector *gainCorrector = 0;
	gainCorrector = calServer.getAtom(gainCorrector);

	if (nodes)
		if (checkDope(nodes)) {
			CalInt8Vector nodesVec(int8ptr(nodes->addr), vecSize);

			gainCorrector->gainCorrect(timeVec, nodesVec, eInVec, eOutVec);
		} else
			errstr<<"nodes vector is empty"
				  <<Sas::fatal("EmptyVector");
	else
		gainCorrector->gainCorrect(timeVec, eInVec, eOutVec);
} // CAL_gainCorrect()


//
//	-----------------------( CAL_piToeV )-----------------------------------
//
void CXXFUNC(cal_pitoev, CAL_piToeV)(const Dope1 *eIn, Dope1 *eOut)
{
	checkDope(eIn);
	unsigned vecSize = eIn->dim[0].extent;

	real32 *rp = allocVec<real32>(eOut, vecSize);
	CalReal32Vector	eInVec(calServer.real32Vector(FLOATP(eIn->addr),vecSize)),
					eOutVec(calServer.real32Vector(rp, vecSize));

	//
	//	get the atom
	//
	EnergyCorrector *eCorrector = 0;
	eCorrector = calServer.getAtom(eCorrector);

	eCorrector->piToeV(eInVec, eOutVec);
} // CAL_piToeV()



//
//	-------------------( CAL_pnGainCorrect )---------------------------------
//
void CXXFUNC(cal_pngaincorrect, CAL_pnGainCorrect)
							(const Dope1 *rawX,
							 const Dope1 *rawY,
							 const Dope1 *eIn,
							 Dope1 *eOut)
{
	checkDope(eIn);
	unsigned vecSize = eIn->dim[0].extent;

	real32 *rp = allocVec<real32>(eOut, vecSize);
	double dum; CalReal64Vector timeVec(&dum, 1);
	CalReal32Vector	eInVec(calServer.real32Vector(FLOATP(eIn->addr),vecSize)),
					eOutVec(calServer.real32Vector(rp, vecSize));
	CalInt16Vector	rawXVec(SHORTP(rawX->addr), vecSize),
					rawYVec(SHORTP(rawY->addr), vecSize);

	//
	//	get the atom
	//
	EnergyCorrector *gainCorrector = 0;
	gainCorrector = calServer.getAtom(gainCorrector);

	gainCorrector->gainCorrect(timeVec, rawXVec, rawYVec, eInVec, eOutVec);
} // CAL_pnGainCorrect()


//
//	-------------------( CAL_mosGainCorrect )---------------------------------
//
void CXXFUNC(cal_mosgaincorrect, CAL_mosGainCorrect)
							(const Dope1 *time,
							 const Dope1 *pattern,
							 const Dope1 *eIn,
							 Dope1 *eOut)
{
	checkDope(eIn);
	unsigned vecSize = eIn->dim[0].extent;

	real32 *rp = allocVec<real32>(eOut, vecSize);
	checkDope(time);
	checkDope(pattern);
	CalReal64Vector timeVec(DOUBLEP(time->addr), vecSize);
	CalReal32Vector	eInVec(calServer.real32Vector(FLOATP(eIn->addr),vecSize)),
					eOutVec(calServer.real32Vector(rp, vecSize));
	CalInt8Vector	patternVec(int8ptr(pattern->addr), vecSize);

	//
	//	get the atom
	//
	EnergyCorrector *gainCorrector = 0;
	gainCorrector = calServer.getAtom(gainCorrector);

	gainCorrector->gainCorrect(timeVec, eInVec, patternVec, eOutVec);
} // CAL_mosGainCorrect()


//
//	-------------------( CAL_mosCtiCorrect )---------------------------------
//
void CXXFUNC(cal_moscticorrect, CAL_mosCtiCorrect)
							(const Dope1 *time,
							 const Dope1 *rawX,
							 const Dope1 *rawY,
							 const Dope1 *eIn,
							 Dope1 *eOut)
{
	checkDope(time);
	checkDope(rawX);
	checkDope(rawY);
	checkDope(eIn);
	unsigned vecSize = eIn->dim[0].extent;

	real32 *rp = allocVec<real32>(eOut, vecSize);
	CalReal64Vector timeVec(DOUBLEP(time->addr), vecSize);
	CalInt16Vector	rawXVec(SHORTP(rawX->addr), vecSize),
					rawYVec(SHORTP(rawY->addr), vecSize);
	CalReal32Vector	eInVec(calServer.real32Vector(FLOATP(eIn->addr),vecSize)),
					eOutVec(calServer.real32Vector(rp, vecSize));

	//
	//	do it
	//
	CtiCorrector *ctiCorrector = 0;
	ctiCorrector = calServer.getAtom(ctiCorrector);
	ctiCorrector->ctiCorrect(timeVec, rawXVec, rawYVec, eInVec, eOutVec);
} // CAL_mosCtiCorrect()


//
//	-------------------( CAL_pnCtiCorrect )--------------------------------
//
void CXXFUNC(cal_pncticorrect, CAL_pnCtiCorrect)
							(const Dope1 *rawX,
							 const Dope1 *rawY,
							 const Dope1 *frame,
							 const Dope1 *eIn,
							 Dope1 *eOut)
{
	checkDope(rawX);
	checkDope(rawY);
	checkDope(frame);
	checkDope(eIn);
	unsigned vecSize = eIn->dim[0].extent;

	real32 *rp = allocVec<real32>(eOut, vecSize);
	double dum; CalReal64Vector timeVec(&dum, 1);
	CalInt16Vector	rawXVec(SHORTP(rawX->addr), vecSize),
					rawYVec(SHORTP(rawY->addr), vecSize);
	CalInt32Vector  frameVec(INTP(frame->addr), vecSize);
	CalReal32Vector	eInVec(calServer.real32Vector(FLOATP(eIn->addr),vecSize)),
					eOutVec(calServer.real32Vector(rp, vecSize));

	//
	//	do it
	//
	CtiCorrector *ctiCorrector = 0;
	ctiCorrector = calServer.getAtom(ctiCorrector);
	ctiCorrector->ctiCorrect(timeVec, rawXVec, rawYVec, frameVec,
							 eInVec, eOutVec);
} // CAL_pnCtiCorrect()


Double CXXFUNC(cal_pnreemissionthreshold, CAL_pnReEmissionThreshold)
							(const Double *energy,	//	[ADU]
							 const Integer2 *rawY)
{
	assert(energy && rawY);
	EnergyCorrector *ec = 0; ec = calServer.getAtom(ec);

	return  ec->reEmissionThreshold(*energy, *rawY);
} // CAL_pnReEmissionThreshold()

void CXXFUNC(cal_pnadditionalgaincticorrect, CAL_pnAdditionalGainCtiCorrect)
							(Dope1 *eIn,
							 const Dope1 *rawY,
							 const Double *T,
							 Dope1 *parOut,
							 const Logical *doTCorr,
							 const Logical *doOffsetCor,
							 const Logical *doLongTermCTISoftCorr,
							 const Logical *doLongTermCTIyCorr,
							 const Logical *doLongTermCTICorr)
{
	checkDope(eIn);
	checkDope(rawY);
	unsigned N = eIn->dim[0].extent;
	CalReal32Vector eInVec(FLOATP(eIn->addr), N);
	CalInt16Vector rawYvec(SHORTP(rawY->addr), N);
	assert(T);
	bool sw[] = { doTCorr ? *doTCorr : true,
				  doOffsetCor ? *doOffsetCor : true,
				  doLongTermCTISoftCorr ? *doLongTermCTISoftCorr : true,
				  doLongTermCTIyCorr ? *doLongTermCTIyCorr : true,
				  doLongTermCTICorr ? *doLongTermCTICorr : true };

	EnergyCorrector *ecorr = 0; ecorr = calServer.getAtom(ecorr);
	vector<double> v = ecorr->additionalGainCtiCorrect(eInVec, rawYvec, *T,
											sw[0], sw[1], sw[2], sw[3], sw[4]);

	real32 *outv = allocVec<real32>(parOut, v.size());
	vector2cvec(v, outv, v.size());
} // CAL_pnAdditionalGainCtiCorrect()

void CXXFUNC(cal_pngaintimingcorrect, CAL_pnGainTimingCorrect)(Dope1 *eIn)
{
	checkDope(eIn);
	unsigned N = eIn->dim[0].extent;
	CalReal32Vector eInVec(FLOATP(eIn->addr), N);

	EnergyCorrector *ecorr = 0; ecorr = calServer.getAtom(ecorr);
	ecorr->gainTimingCorrect(eInVec);
} // CAL_pnGainTimingCorrect()

void CXXFUNC(cal_pngainburstcorrect, CAL_pnGainBurstCorrect)(Dope1 *eIn)
{
	checkDope(eIn);
	unsigned N = eIn->dim[0].extent;
	CalReal32Vector eInVec(FLOATP(eIn->addr), N);

	EnergyCorrector *ecorr = 0; ecorr = calServer.getAtom(ecorr);
	ecorr->gainBurstCorrect(eInVec);
} // CAL_pnGainBurstCorrect()

void CXXFUNC(cal_pngaineffcorrect, CAL_pnGainEffCorrect)(Dope1 *eIn)
{
	checkDope(eIn);
	unsigned N = eIn->dim[0].extent;
	CalReal32Vector eInVec(FLOATP(eIn->addr), N);

	EnergyCorrector *ecorr = 0; ecorr = calServer.getAtom(ecorr);
	ecorr->gainEffCorrect(eInVec);
} // CAL_pnGainEffCorrect()

void CXXFUNC(cal_pngainffcorrect, CAL_pnGainFfCorrect)(Dope1 *eIn)
{
	checkDope(eIn);
	unsigned N = eIn->dim[0].extent;
	CalReal32Vector eInVec(FLOATP(eIn->addr), N);

	EnergyCorrector *ecorr = 0; ecorr = calServer.getAtom(ecorr);
	ecorr->gainFfCorrect(eInVec);
} // CAL_pnGainFfCorrect()

//
//	-------------------( CAL_pnMedianMap )---------------------------------
//
void CXXFUNC(cal_pnmedianmap, CAL_pnMedianMap)
							(Dope2 *medianMap)
{
	//
	//	get the atom
	//
	LowEnergyNoiseServer *lens = 0; lens = calServer.getAtom(lens);

  // Read in the median map for the current mode, ccd combination
	const SImage &map = lens->medianMap();
  Single *data = new Single[map.size()];

  unsigned n=0,
           NY = map.xExtent(),  // Bizarre but xExtent = rows &
           NX = map.yExtent();  // yextent = columns
           
  errstr<<"Found "<<NX<<","<<NY<<" elements in median map "
          << Sas::message(Msg::App, Msg::Noisy);
          
  for (unsigned nx=0; nx<NX; ++nx) {
      for (unsigned ny=0; ny<NY; ++ny) {
            data[n++] = map[ny][nx];    // Is this the right way round ?
      }
  }

  // setup dope vectors
  vector<unsigned> twoDim(2); twoDim[0] = NX;
  twoDim[1] = NY;
  setupDope(*medianMap, data, twoDim);         

  // release the Atom because it takes up a lot of memory
  calServer.releaseAtom(lens);
  
} // CAL_pnMedianMap()

//
//	-------------------( CAL_pnOffsetCorrValues )---------------------------------
//
void CXXFUNC(cal_pnoffsetcorrvalues, CAL_pnOffsetCorrValues)
							(Dope1 *hlo,
               Dope1 *hhi,
               Dope1 *val)
{
	//
	//	get the atom
	//
	LowEnergyNoiseServer *lens = 0; lens = calServer.getAtom(lens);

  // Get the correction energies and values for the current mode, ccd combination
	const vector<real32> &lowen = lens->lowEnergy();
  const vector<real32> &highen = lens->highEnergy();
  const vector<int16>  &corrVal = lens->correctionValue();

  // Move to local array
  unsigned N = lowen.size();           
  Single *lowenData = new Single[N];
  Single *highenData = new Single[N];
  //Integer2 *corrValData = new Integer2[N];
  //int16 *corrValData = new int16[N];
  int16 *corrValData = allocVec<int16>(val, N);
  
  for (unsigned n=0; n<N; ++n) {
      lowenData[n] = lowen[n];                                    
      highenData[n] = highen[n];
      corrValData[n] = corrVal[n];
  }

  // setup dope vectors 
  vector<unsigned> oneDim(1); oneDim[0] = N;
  setupDope(*hlo, lowenData, oneDim);
  setupDope(*hhi, highenData, oneDim);
  //setupDope(*val, corrValData, oneDim);
  
  // release the Atom because it takes up a lot of memory
  calServer.releaseAtom(lens);

} // CAL_pnOffsetCorrValues()
                                                           
//
//	-------------------( CAL_pnNoiseMap )---------------------------------
//
void CXXFUNC(cal_pnnoisemap, CAL_pnNoiseMap)
							(Dope1 *rawX,
               Dope3 *noiseMap)
{
	//
	//	get the atom
	//
	LowEnergyNoiseServer *lens = 0; lens = calServer.getAtom(lens);

  // Get the rawX value vector
  const vector<int8> &rawxvector = lens->rawxVector();
  unsigned N = rawxvector.size();
  Integer2 *rawxvecData = new Integer2[N];
  
  // Read in the median map for the current mode, ccd combination
	const vector<SImage> &mapvector = lens->noiseMapVector();
  unsigned NY = mapvector[0].xExtent(),
           NX = mapvector[0].yExtent();
  Single *mapvecData = new Single[N*NY*NX];
  
  // setup dope vectors
  vector<unsigned> oneDim(1); oneDim[0] = N;
  vector<unsigned> threeDim(3); threeDim[0] = oneDim[0];
  threeDim[1] = NX; threeDim[2] = NY;
    
  setupDope(*rawX, rawxvecData, oneDim);
  setupDope(*noiseMap, mapvecData, threeDim);

  // Move data into doped arrays
  Integer3 nmap=0;
  for (unsigned n=0; n<N; ++n) {
      rawxvecData[n] = static_cast<Integer2>(rawxvector[n]);
                     
      for (unsigned nx=0; nx<NX; ++nx) {
        for (unsigned ny=0; ny<NY; ++ny) {
            mapvecData[nmap++] = mapvector[n][ny][nx];    // Is this the right way round ?
        }
      }
  }
  
  // release the Atom because it takes up a lot of memory
  calServer.releaseAtom(lens);

} // CAL_pnMedianMap()
void CXXFUNC(cal_rgsoffsetcorrect, CAL_rgsOffsetCorrect)
							(const Dope1 *time,
							 const Dope1 *rawX,
							 const Dope1 *rawY,                       
							 const Dope1 *shapes,
							 const Dope1 *nodes,
							 const Dope1 *eIn,              
							 Dope1 *eOut)
{
	checkDope(time);
	checkDope(rawX);
	checkDope(rawY);
	checkDope(shapes);
	checkDope(nodes);
	checkDope(eIn);
	unsigned vecSize = eIn->dim[0].extent;

	real32 *rp = allocVec<real32>(eOut, vecSize);
	const CalReal64Vector timeVec(DOUBLEP(time->addr), vecSize);
	const CalInt16Vector  rawXVec(SHORTP(rawX->addr), vecSize),
						  rawYVec(SHORTP(rawY->addr), vecSize);
	const CalInt8Vector	shapeVec(int8ptr(shapes->addr), vecSize),
						nodesVec(int8ptr(nodes->addr), vecSize);
	const CalInt16Vector
					eInVec(calServer.int16Vector(SHORTP(eIn->addr),vecSize));
	CalReal32Vector eOutVec(calServer.real32Vector(rp, vecSize));

	//
	//	get the atom
	//
	EnergyCorrector *offsetCorrector = 0;
	offsetCorrector = calServer.getAtom(offsetCorrector);

	offsetCorrector->offsetCorrect(timeVec, rawXVec, rawYVec, shapeVec,
								   nodesVec, eInVec, eOutVec);
} // CAL_rgsOffsetCorrect()



void CXXFUNC(cal_rgscticorrect, CAL_rgsCtiCorrect)
							(const Dope1 *time,
							 const Dope1 *rawX,
							 const Dope1 *rawY,
							 const Dope1 *nodes,
							 const Dope1 *shape,
							 const Dope1 *separation,
							 const Dope1 *eIn,
							 Dope1 *eOut)
{
	checkDope(time);
	checkDope(rawX);
	checkDope(rawY);
	checkDope(nodes);
	checkDope(shape);
	checkDope(separation);
	checkDope(eIn);
	unsigned vecSize = eIn->dim[0].extent;

	real32 *rp = allocVec<real32>(eOut, vecSize);
	CalReal64Vector timeVec(DOUBLEP(time->addr), vecSize);
	CalInt8Vector	nodesVec(int8ptr(nodes->addr), vecSize),
					shapeVec(int8ptr(shape->addr), vecSize);
	CalInt16Vector	rawXVec(SHORTP(rawX->addr), vecSize),
					rawYVec(SHORTP(rawY->addr), vecSize),
				    sepVec(SHORTP(separation->addr), vecSize);
	CalReal32Vector	eInVec(calServer.real32Vector(FLOATP(eIn->addr),vecSize)),
					eOutVec(calServer.real32Vector(rp, vecSize));

	//
	//	do it
	//
	CtiCorrector *ctiCorrector = 0;
	ctiCorrector = calServer.getAtom(ctiCorrector);
	ctiCorrector->ctiCorrect(timeVec, rawXVec, rawYVec, shapeVec, sepVec,
							 eInVec, eOutVec, &nodesVec);
} // CAL_rgsCtiCorrect()


//
// ---------------------( CAL_rgsGetCrossPSF )--------------------------------
//
void CXXFUNC(cal_rgsgetcrosspsf, CAL_rgsGetCrossPSF)
	                       (const Double *time,
							const Double *beta,
							const Double *sourceXdisp,
							const Dope1 *xAngles,
							const Double *binWidth,
							Dope1 *distribution)
{
	if (!isValidPointer(time) ||
		!isValidPointer(beta) ||
		!isValidPointer(sourceXdisp)||
		!isValidPointer(xAngles)||
		!isValidPointer(binWidth))
		errstr<<"time/beta/sourceXdisp/xAngles/binWidth argument(s) to "
				"rgsCrossPSF construction are void"
			  <<Sas::error("ArgumentMissing");

	unsigned vecSize = xAngles->dim[0].extent;	
	real64 *rp = allocVec<real64>(distribution, vecSize);
	CalReal64Vector	xAnglesVec(DOUBLEP(xAngles->addr), vecSize),
					distVec(rp, vecSize);

	//	do it
	CanonicalCrossPsf *cpsf = 0; cpsf = calServer.getAtom(cpsf);
	Spectrum prob(cpsf->probability(*time, *beta, *sourceXdisp,
				  xAnglesVec, *binWidth));

	unsigned N = prob.size();
	for (unsigned n=0; n<N; ++n)
		distVec[n] = prob[n];
} // CAL_rgsGetCrossPSF


//
//	--------------------( CAL_rgsGetEventThresholds )--------------------------
//
void CXXFUNC(cal_rgsgeteventthresholds, CAL_rgsGetEventThresholds)
							(Dope1 *rejectionThresholds,
							 Dope1 *acceptanceThresholds,
							 const Integer1 *node)
{
	EnergyCorrector *ec = 0; ec = calServer.getAtom(ec);
	unsigned N = ec->nGrades();
	real32 *accThresh = allocVec<real32>(rejectionThresholds, N),
		   *rejThresh = allocVec<real32>(acceptanceThresholds, N);
	CalReal32Vector accVec(accThresh, N),
					rejVec(rejThresh, N);

	Node nodeArg = calServer.state()->node()->get();
	if (node)
		//	node argument given
		nodeArg = Xmm::toNode(*node);

	for (unsigned n=0; n<ec->nGrades(); ++n) {
		accVec[n] = ec->acceptanceThreshold(n+1, nodeArg);
		rejVec[n] = ec->rejectionThreshold(n+1, nodeArg);
	}
} // CAL_rgsGetEventThresholds()



//
//	--------------------------( CAL_rgsGetLSF )------------------------------
//
void CXXFUNC(cal_rgsgetlsf, CAL_rgsGetLSF)
				(const Double *lambda, const Double *theta, const Double *phi,
				 const Integer3 *diffOrder,
				 const Double *lambdaMin, const Double *lambdaMax,
				 Dope1 *lambdaGrid, Dope1 *lsfProf)
{
	assert(lambda && theta && phi && diffOrder && lambdaMin && lambdaMax);
	errstr<<"Construction of RGS LSF not implemented yet"
		  <<Sas::warning("NotImplementedYet");

	Double *lgridVec = new Double[2];
	lgridVec[0] = *lambdaMin;
	lgridVec[1] = *lambdaMax;

	Double *lsfprofVec = new Double[2];
	lsfprofVec[0] = lsfprofVec[1] = 0.;

	//	setup dope vectors
	vector<unsigned>	oneDim(1); oneDim[0] = 2;
	setupDope(*lambdaGrid, lgridVec, oneDim);
	setupDope(*lsfProf, lsfprofVec, oneDim);
} // CAL_rgsGetLSF()


//
//	--------------------( CAL_rgsGetRedistribution )------------------------
//
void CXXFUNC(cal_rgsgetredistribution, CAL_rgsGetRedistribution)
					(const Double *E,
					 const Integer2 *rawX,
					 const Integer2 *rawY,
					 const Double *, // rejectionThreshold - ignored
					 Dope1 *redistribution,
					 Dope1 *channels)
{
	CanonicalRedist *redist = 0; redist = calServer.getAtom(redist);
	Spectrum spec(redist->at(*rawX, *rawY).spectrum(*E));

	unsigned N = spec.size();

	Double *vec = allocVec<Double>(redistribution, N);
	vector2cvec(spec, vec, N);

	if (channels) {
		Integer2 *vec = allocVec<Integer2>(channels, N);
		for (unsigned n=0; n<N; ++n)
			vec[n] = SHORT(n);
	}
} // CAL_rgsGetRedistribution()


//
//	--------------------( CAL_rgsGetRefinedWaveScale )------------------------
//
void CXXFUNC(cal_rgsgetrefinedwavescale, CAL_rgsGetRefinedWaveScale)
				(const Double *time, const Dope1 *order,
				 Dope1 *wavelength)
{
	assert(time && order && wavelength);
} // CAL_rgsGetRefinedWaveScale()


//
//	OM specific
//

//
//	---------------------( CAL_omDistortion )-------------------------------
//
void CXXFUNC(cal_omdistortion, CAL_omDistortion)
				(const Dope1 *rawX, const Dope1 *rawY,
				 Dope1 *deltaX, Dope1 *deltaY, Dope1 *sigmaX, Dope1 *sigmaY,
				 const Integer3 *calcMode, const Integer3 *nInterPoints)
{
	checkDope(rawX);
	checkDope(rawY);
	unsigned vecSize = rawX->dim[0].extent;

	Single *delX = allocVec<Single>(deltaX, vecSize),
		   *delY = allocVec<Single>(deltaY, vecSize),
		   *sigX = 0,
		   *sigY = 0;
	if (sigmaX && sigmaY) {
		sigX = allocVec<Single>(sigmaX, vecSize);
		sigY = allocVec<Single>(sigmaY, vecSize);
	}

	CalReal32Vector	rawXvec(FLOATP(rawX->addr), vecSize),
					rawYvec(FLOATP(rawY->addr), vecSize),
					delXvec(delX, vecSize),
					delYvec(delY, vecSize),
					sigXvec(sigX, vecSize),
					sigYvec(sigY, vecSize);

	//
	//	get the atom
	//
	OmGeoLinearizer *geoLin = 0; geoLin = calServer.getAtom(geoLin);

	if (nInterPoints)
		//	set number of points used in interpolation
		geoLin->additionalPoints(*nInterPoints);

	unsigned mode = 0;
	if (calcMode) {
		if (*calcMode != 0 && *calcMode != 1)
			errstr<<"calcMode must be either 0 or 1 - got: "<<*calcMode
				  <<Sas::error("IllegalInput");
		mode = *calcMode;
	}

	if (!mode)
		geoLin->interpolatedDeltas(rawXvec, rawYvec, delXvec, delYvec,
								   sigXvec, sigYvec);
	else
		geoLin->polynomialDeltas(rawXvec, rawYvec, delXvec, delYvec,
								 sigXvec, sigYvec);
} // CAL_omDistortion()

//
//	---------------------( CAL_omInverseDistortion )-------------------------------
//
void CXXFUNC(cal_ominversedistortion, CAL_omInverseDistortion)
				(const Dope1 *rawX, const Dope1 *rawY,
				 Dope1 *deltaX, Dope1 *deltaY, Dope1 *sigmaX, Dope1 *sigmaY,
				 const Integer3 *calcMode, const Integer3 *nInterPoints)
{
	checkDope(rawX);
	checkDope(rawY);
	unsigned vecSize = rawX->dim[0].extent;

	Single *delX = allocVec<Single>(deltaX, vecSize),
		   *delY = allocVec<Single>(deltaY, vecSize),
		   *sigX = 0,
		   *sigY = 0;
	if (sigmaX && sigmaY) {
		sigX = allocVec<Single>(sigmaX, vecSize);
		sigY = allocVec<Single>(sigmaY, vecSize);
	}

	CalReal32Vector	rawXvec(FLOATP(rawX->addr), vecSize),
					rawYvec(FLOATP(rawY->addr), vecSize),
					delXvec(delX, vecSize),
					delYvec(delY, vecSize),
					sigXvec(sigX, vecSize),
					sigYvec(sigY, vecSize);

	//
	//	get the atom
	//
	OmGeoLinearizer *geoLin = 0; geoLin = calServer.getAtom(geoLin);

	if (nInterPoints)
		//	set number of points used in interpolation
		geoLin->additionalPoints(*nInterPoints);

	unsigned mode = 0;
	if (calcMode) {
		if (*calcMode != 0 && *calcMode != 1)
			errstr<<"calcMode must be either 0 or 1 - got: "<<*calcMode
				  <<Sas::error("IllegalInput");
		mode = *calcMode;
	}

	if (!mode)
		geoLin->interpolatedDeltas(rawXvec, rawYvec, delXvec, delYvec,
								   sigXvec, sigYvec);
	else
		geoLin->inversepolynomialDeltas(rawXvec, rawYvec, delXvec, delYvec,
								 sigXvec, sigYvec);
} // CAL_omInverseDistortion()

//
//	---------------( CAL_omGetGrismFlux/WavelengthColorTransform )------------
//
void CXXFUNC(cal_omgetgrismflux, CAL_omGetGrismFlux)
						   	 (Dope1 *lambdaVec,
							  Dope1 *isfVec,
						   	  Dope1 *errorVec)
{
	OmGrismDataServer *gds = 0; gds = calServer.getAtom(gds);
	const OmGrismDataServer::ISF &isfData = gds->isfData();

	unsigned N = isfData.wavelength.size();
	Double *lambdaIsf = new Double [N],
		   *isf = new Double [N],
		   *isfError = new Double [N];

	vector2cvec(isfData.wavelength, lambdaIsf, N);
	vector2cvec(isfData.data, isf, N);
	vector2cvec(isfData.error, isfError, N);

	vector<unsigned> oneDim(1); oneDim[0] = N;
	setupDope(*lambdaVec, lambdaIsf, oneDim);
	setupDope(*isfVec, isf, oneDim);
	setupDope(*errorVec, isfError, oneDim);
} // CAL_omGetGrismFlux()


void CXXFUNC(cal_omgetgrismwavelength, CAL_omGetGrismWavelength)
						 (Dope1 *lambdaVec)
{
	OmGrismDataServer *gds = 0; gds = calServer.getAtom(gds);
	const vector<double> &lambda = gds->wavelength();

	unsigned N = lambda.size();
	Double *l = new Double[N];
	vector2cvec(lambda, l, N);
	vector<unsigned> oneDim(1); oneDim[0] = N;
	setupDope(*lambdaVec, l, oneDim);
} // CAL_omGetGrismWavelength()


//
//	------------------------( CAL_omColorTransform )-------------------------
//
Handle<OmColorTransformator> CXXFUNC(cal_omgetcolortransformator,
									 CAL_omGetColorTransformator)
							(const Integer3 *filter1,
							 const Integer3 *filter2)
{
	assert(filter1 && filter2);

	//
	//	get the atom
	//
	OmColorTransformator *ct = 0; ct = calServer.getAtom(ct);
	Filter f1 = static_cast<Filter>(*filter1),
		   f2 = static_cast<Filter>(*filter2);

	ct->withFilters(f1, f2);

	return Handle<OmColorTransformator>(ct);
} // CAL_omGetColorTransformator()


Integer1 CXXFUNC(cal_omcolortransbranches, CAL_omColorTransBranches)
						(const Handle<OmColorTransformator> *colorTrans)
{
	return colorTrans->addressOfObject()->branches();
} // CAL_omColorTransBranches()


void CXXFUNC(cal_omcolortranssetbranch,	CAL_omColorTransSetBranch)
							(const Handle<OmColorTransformator> *colorTrans,
							 const Integer1 *branch)
{
	assert(branch);
	colorTrans->addressOfObject()->branch(*branch);
} // CAL_omColorTransSetBranch()


Integer1 CXXFUNC(cal_omcolortransvalidityranges,
				 CAL_omColorTransValidityRanges)
							(const Handle<OmColorTransformator> *colorTrans)
{
	assert(colorTrans);
	return colorTrans->addressOfObject()->validityRanges().size();
} // CAL_omColorTransValidityRanges()


IntervalT CXXFUNC(cal_omcolortransvalidityrange, CAL_omColorTransValidityRange)
							(const Handle<OmColorTransformator> *colorTrans,
							 const Integer1 *n)
{
	assert(colorTrans);
	Interval<double> i0 =
		colorTrans->addressOfObject()->validityRanges()[n ? *n : 0];
	IntervalT i1;
	i1.lower = i0.lo();
	i1.upper = i0.hi();

	return i1;
} // CAL_omColorTransValidityRange()


ValueErrorT CXXFUNC(cal_omstandardcolor, CAL_omStandardColor)
							(const Handle<OmColorTransformator> *colorTrans,
							 const Double *magnitude1,
							 const Double *magnitude2)
{
	assert(colorTrans && magnitude1 && magnitude2);
	ValueError ve0 = colorTrans->addressOfObject()->standardColor(*magnitude1,
																  *magnitude2);
	ValueErrorT ve1;
	ve1.value = ve0.value;
	ve1.error = ve0.error;

	return ve1;
} // CAL_omStandardColor()


ValueErrorT CXXFUNC(cal_omstandardmagnitude, CAL_omStandardMagnitude)
							(const Handle<OmColorTransformator> *colorTrans,
							 const ValueErrorT *stdColor,
							 const Double *magnitude)
{
	assert(colorTrans && stdColor && magnitude);
	ValueError stdcolve;
	stdcolve.value = stdColor->value;
	stdcolve.error = stdColor->error;
	ValueError ve0 = colorTrans->addressOfObject()->
									standardMagnitude(stdcolve, *magnitude);

	ValueErrorT ve1;
	ve1.value = ve0.value;
	ve1.error = ve0.error;

	return ve1;
} // CAL_omStandardMagnitude()


void CXXFUNC(cal_omcolortransform, CAL_omColorTransform)
							(const Integer3 *filter1,
							 const Integer3 *filter2,
							 const Double *magnitude1,
							 const Double *magnitude2,
							 Double *stdColor12,
							 Double *stdMagnitude2,
							 Double *stdColor12Error,
							 Double *stdMagnitude2Error)
{
	//
	//	get the atom
	//
	OmColorTransformator *ct = 0; ct = calServer.getAtom(ct);
	Filter f1 = static_cast<Filter>(*filter1),
		   f2 = static_cast<Filter>(*filter2);

	ct->withFilters(f1, f2);
	ValueError col = ct->standardColor(*magnitude1, *magnitude2),
			   mag = ct->standardMagnitude(col, *magnitude2);

	*stdColor12 = col.value;
	*stdMagnitude2 = mag.value;

	if (stdColor12Error)
		*stdColor12Error = col.error;
	if (stdMagnitude2Error)
		*stdMagnitude2Error = mag.error;
} // CAL_omColorTransform()


//
//	----------------------( CAL_omGetApertureRadius )------------------------
//
Double CXXFUNC(cal_omgetapertureradius, CAL_omGetApertureRadius)()
{
	OmColorTransformator *ct = 0; ct = calServer.getAtom(ct);
	OmPsf *psf = 0; psf = calServer.getAtom(psf);

	double radius0 = ct->apertureRadius(),
		   radius1 = psf->apertureRadius();
	if (radius0 != radius1)
		errstr<<"The aperture radius value from CCF::OM_COLORTRANS ("
			  <<radius0<<") for filter "<<calServer.state()->filter()->get()
			  <<" does not coincide with the corresponding value from "
				"CCF::OM_PHOTONAT ("<<radius1<<") - this is a CCF problem."
			  <<Sas::error("CcfIntegrityError");

	return radius0;
} // CAL_omGetApertureRadius()


//
//	----------------------( CAL_omGetFrameParameters )----------------------
//
void CXXFUNC(cal_omgetframeparameters, CAL_omGetFrameParameters)
				(struct OMframeParameters *frameParam)
{
	//
	//	get the atom
	//
	OmFluxConverter *fc = 0; fc = calServer.getAtom(fc);
	OmFluxConverter::FrameParameters fp = fc->frameParameters();

	frameParam->nHorizontalPixels = fp.nHorizontalPixels.value;
    frameParam->nVerticalPixels = fp.nVerticalPixels.value;
	frameParam->verticalTransferTime = fp.verticalTransferTime.value;
	frameParam->horizontalTransferTime = fp.horizontalTransferTime.value;
	frameParam->nVerticalPixelsImage = fp.nVerticalPixelsImage.value;
	frameParam->verticalHorizontalSwitchTime =
									fp.verticalHorizontalSwitchTime.value;
	frameParam->horizontalVerticalSwitchTime =
									fp.horizontalVerticalSwitchTime.value;
	frameParam->redundantCcdOffset = fp.redundantCcdOffset.value;
} // CAL_omGetFrameParameters()


//
//	---( OmPsf )-------------------------------------------------------------
//
Handle<OmPsf> CXXFUNC(cal_omgetpsf, CAL_omGetPSF)
							(const Double *cenX,
							 const Double *cenY,
							 const Double *countFrameRateRatio)
{
	assert(cenX && cenY && countFrameRateRatio);
	OmPsf *psf = 0; psf = calServer.getAtom(psf);
	psf->forCountFrameRateRatio(*countFrameRateRatio).withCenterAt(
																*cenX, *cenY);

	return Handle<OmPsf>(psf);
} // CAL_omGetPSF()


Double CXXFUNC(cal_ompsfencircledenergy, CAL_omPsfEncircledEnergy)
							(const Handle<OmPsf> *psf,
							 const Double *radius)
{
	assert(psf && radius);

	return psf->addressOfObject()->encircledEnergy()(*radius);
} // CAL_omPsfEncircledEnergy()


Double CXXFUNC(cal_ompsfcircleradius, CAL_omPsfCircleRadius)
							(const Handle<OmPsf> *psf,
							 const Double *fraction)
{
	assert(psf && fraction);
	return psf->addressOfObject()->circleRadius()(*fraction);
} // CAL_omPsfCircleRadius()


void CXXFUNC(cal_omgetpsfmap, CAL_omGetPSFmap)
				(const Double *rawX, const Double *rawY,
				 const Integer3 *nPixelsNegX, const Integer3 *nPixelsPosX,
				 const Integer3 *nPixelsNegY, const Integer3 *nPixelsPosY,
				 const Integer3 *xRes, const Integer3 *yRes,
				 const Double *countFrameRateRatio,
				 Dope2 *PSFmap)
{
	OmPsf *psf = 0; psf = calServer.getAtom(psf);

	OmPsf::OmPsfWindow win = { *nPixelsNegX, *nPixelsPosX,
							   *nPixelsNegY, *nPixelsPosY };
	SImage map(psf->forCountFrameRateRatio(*countFrameRateRatio).
					withCenterAt(*rawX, *rawY).image(win, *xRes, *yRes));

	Single *mapData = new Single[map.size()];
	memcpy(mapData, map.data(), map.memSize());

	//	setup dope vectors
	vector<unsigned> twoDim(2); twoDim[0] = map.xExtent();
								twoDim[1] = map.yExtent();
	setupDope(*PSFmap, mapData, twoDim);
} // CAL_omGetPSFmap()


//
//	----------------------( CAL_omGetPlateScale )---------------------------
//
void CXXFUNC(cal_omgetplatescale, CAL_omGetPlateScale)
				(const Double *lambda,
				 Double *plateScale, Double *errorX, Double *errorY)
{
	assert(lambda);
	OmGeoLinearizer *geoLin = 0; geoLin = calServer.getAtom(geoLin);
	*plateScale = geoLin->plateScale(*lambda);
	if (errorX && errorY)
		*errorX = *errorY = 0.;
} // CAL_omGetPlateScale()


//
//	--------------------( CAL_omLargeSenseVariation )-----------------------
//
void CXXFUNC(cal_omlargesensevariation, CAL_omLargeSenseVariation)
					(const Integer3 *x0, const Integer3 *y0,
					 const Integer3 *xSize, const Integer3 *ySize,
					 const Double *lambda,
					 Dope2 *senseMap, Double *averageFlatField,
					 Double *averageFlatFieldError,
					 Double *pixelError)
{
	//	get the atom
	OmFlatFieldServer *ffs = 0; ffs = calServer.getAtom(ffs);

	//	get pointer to the flatfield
	const SImage *map = ffs->largeScaleFlatFieldMap(UNT(*x0), UNT(*y0),
												    UNT(*xSize), UNT(*ySize),
													*lambda);

    //  setup dope vectors
    vector<unsigned> twoDim(2); twoDim[0] = *xSize; twoDim[1] = *ySize;
    setupDope(*senseMap, map->data(), twoDim);

	if (averageFlatField && averageFlatFieldError) {
		ValueError ve = ffs->largeScaleFlatFieldAverage();
		*averageFlatField = ve.value;
		*averageFlatFieldError = ve.error;
	}
	if (pixelError)
		*pixelError = ffs->largeScalePixelError();

	//	release the Atom because it takes up a lot of memory
	calServer.releaseAtom(ffs);
} // CAL_omLargeSenseVariation()


//
//	--------------------( CAL_omLEDtemplateMap )---------------------
//
void CXXFUNC(cal_omledtemplatemap, CAL_omLEDtemplateMap)
					(const Integer3 *x0, const Integer3 *y0,
					 const Integer3 *xSize, const Integer3 *ySize,
					 Dope2 *outmap, Double *average,
					 Double *averageError,
					 Double *pixelError)
{
	//	get the atom
	OmFlatFieldServer *ffs = 0; ffs = calServer.getAtom(ffs);

	//	get pointer to the flatfield
	const SImage *map = ffs->ledTemplateMap(UNT(*x0), UNT(*y0),
										    UNT(*xSize), UNT(*ySize));

    //  setup dope vectors
    vector<unsigned> twoDim(2); twoDim[0] = *xSize; twoDim[1] = *ySize;
    setupDope(*outmap, map->data(), twoDim);

	if (average && averageError) {
		ValueError ve = ffs->ledTemplateAverage();
		*average = ve.value;
		*averageError = ve.error;
	}
	if (pixelError)
		*pixelError = ffs->ledTemplatePixelError();

	//	release the Atom because it takes up a lot of memory
	calServer.releaseAtom(ffs);
} // CAL_omLEDtemplateMap()



//
//	--------------------( CAL_omPixelSenseVariation )---------------------
//
void CXXFUNC(cal_ompixelsensevariation, CAL_omPixelSenseVariation)
					(const Integer3 *x0, const Integer3 *y0,
					 const Integer3 *xSize, const Integer3 *ySize,
					 const Double *countRateFrameRatio,
					 Dope2 *senseMap, Double *averageFlatField,
					 Double *averageFlatFieldError,
					 Double *pixelError)
{
	//	get the atom
	OmFlatFieldServer *ffs = 0; ffs = calServer.getAtom(ffs);

	//	get pointer to the flatfield
	const SImage *map = ffs->smallScaleFlatFieldMap(UNT(*x0), UNT(*y0),
												    UNT(*xSize), UNT(*ySize),
													*countRateFrameRatio);

    //  setup dope vectors
    vector<unsigned> twoDim(2); twoDim[0] = *xSize; twoDim[1] = *ySize;
    setupDope(*senseMap, map->data(), twoDim);

	if (averageFlatField && averageFlatFieldError) {
		ValueError ve = ffs->smallScaleFlatFieldAverage();
		*averageFlatField = ve.value;
		*averageFlatFieldError = ve.error;
	}
	if (pixelError)
		*pixelError = ffs->smallScalePixelError();

	//	release the Atom because it takes up a lot of memory
	calServer.releaseAtom(ffs);
} // CAL_omPixelSenseVariation()



//
//	------------( CAL_omPhotoMagnitude / CAL_omNaturalMagnitude )--------------
//
void CXXFUNC(cal_omphotomagnitude, CAL_omPhotoMagnitude)
				(const Double *naturalRate, Double *magnitude,
				 const Double *naturalRateErrorMinus,
				 const Double *naturalRateErrorPlus,
				 Double *magnitudeErrorMinus,
				 Double *magnitudeErrorPlus)
{
	OmColorTransformator *ct = 0; ct = calServer.getAtom(ct);
	ValueAsymmetricError vae;
	vae.value = *naturalRate;
	vae.error.errorPlus = 0.;
	vae.error.errorMinus = 0.;
	if (naturalRateErrorMinus)
		vae.error.errorMinus = *naturalRateErrorMinus;
	if (naturalRateErrorPlus)
		vae.error.errorPlus = *naturalRateErrorPlus;

	ValueAsymmetricError mag = ct->toNaturalMagnitudes(vae);
	*magnitude = mag.value;

	if (magnitudeErrorMinus)
		*magnitudeErrorMinus = mag.error.errorMinus;
	if (magnitudeErrorPlus)
		*magnitudeErrorPlus	 = mag.error.errorPlus;
} // CAL_omPhotoMagnitude()


Double CXXFUNC(cal_omnaturalmagnitude, CAL_omNaturalMagnitude)
							(const Double *naturalRate,
							 const Double *naturalRateErrorMinus,
							 const Double *naturalRateErrorPlus,
							 Double *magnitudeErrorMinus,
							 Double *magnitudeErrorPlus)
{
	double mag;
	CXXFUNC(cal_omphotomagnitude, CAL_omPhotoMagnitude)(
				naturalRate, &mag, naturalRateErrorMinus, naturalRateErrorPlus,
				magnitudeErrorMinus, magnitudeErrorPlus);

	return mag;
} // CAL_omNaturalMagnitude()


//
//	-------------------( CAL_omPhotoNatural )----------------------------
//
Double CXXFUNC(cal_omphotonatural0, CAL_omPhotoNatural0)
				(const Double *rawRate, const Double *frameTime,
				 const Double *deadFrac, const Double *eventLoss,
				 const Logical *extendedSource,
				 const Double *rawRateErrorMinus,
				 const Double *rawRateErrorPlus,
				 Double *naturalRateErrorMinus,
				 Double *naturalRateErrorPlus)
{
	//
	//	get the atom
	//
	OmFluxConverter *fc = 0; fc = calServer.getAtom(fc);
	ValueAsymmetricError vain;
	vain.value = *rawRate;
	if (rawRateErrorMinus && rawRateErrorPlus) {
		vain.error.errorPlus = *rawRateErrorPlus;
		vain.error.errorMinus = *rawRateErrorMinus;
	} else
		vain.error.errorPlus = vain.error.errorMinus = 0.;

	ValueAsymmetricError vaout;
	double el = 0.;
	if (eventLoss)
		el = *eventLoss;

	bool extSource = false;
	if (extendedSource)
		extSource = static_cast<bool>(*extendedSource);

	vaout = fc->fluxCorrect(vain, *frameTime, *deadFrac, el, extSource);

	if (naturalRateErrorMinus && naturalRateErrorPlus) {
		*naturalRateErrorMinus = vaout.error.errorMinus;
		*naturalRateErrorPlus = vaout.error.errorPlus;
	}

	return vaout.value;
} // CAL_omPhotoNatural0()


ValueAsymmetricErrorT CXXFUNC(cal_omphotonatural1, CAL_omPhotoNatural1)
				(const ValueAsymmetricErrorT *rawRate, const Double *frameTime,
				 const Double *deadFrac, const Double *eventLoss,
				 const Logical *extendedSource,
				 const Logical *empiricalLinearityCorrection)
{
	//
	//	get the atom
	//
	OmFluxConverter *fc = 0; fc = calServer.getAtom(fc);
	ValueAsymmetricError vain;
	assert(rawRate);
	vain.value = rawRate->value;
	vain.error.errorMinus = rawRate->errorMinus;
	vain.error.errorPlus = rawRate->errorPlus;

	ValueAsymmetricError vaout =
		fc->fluxCorrect(vain, *frameTime, *deadFrac, *eventLoss,
				extendedSource ? static_cast<bool>(*extendedSource) : false,
				empiricalLinearityCorrection ?
					static_cast<bool>(*empiricalLinearityCorrection) : false);

	ValueAsymmetricErrorT v;
	v.value = vaout.value;
	v.errorPlus = vaout.error.errorPlus;
	v.errorMinus = vaout.error.errorMinus;

	return v;
} // CAL_omPhotoNatural1()


//
//	translations from instrument/mode/filter names into integer Ids
//
Integer3 CXXFUNC(instidtoint, instIdToInt)(const char *name, Integer3 length)
{
	string s(f90string(name, length));
	int id = Xmm::toInstrument(s);

	if (id < 0) {
		errstr<<"The string '"<<s<<"' is not among the list of known "
				"instrument identifiers: \n";
		for (unsigned n = 0; n<=Xmm::N_INSTRUMENTS; ++n)
			errstr<<Xmm::toString(static_cast<Xmm::Instrument>(n))<<endl
				  <<Sas::error("Mismatch");
	}

	return id;
} // instIdToInt


Integer3 CXXFUNC(modeidtoint, modeIdToInt)(const char *name, Integer3 length)
{
	string s(f90string(name, length));
	int id = Xmm::toMode(s);

	if (id < 0) {
		errstr<<"The string '"<<s<<"' is not among the list of known "
				"instrument modes: \n";
		for (unsigned n=0; n<=Xmm::N_MODES; ++n)
			errstr<<Xmm::toString(static_cast<Xmm::Mode>(n))<<endl
				  <<Sas::error("Mismatch");
	}

	return id;
} // modeIdToInt


Integer3 CXXFUNC(datamodeidtoint, dataModeIdToInt)
										(const char *name, Integer3 length)
{
	string s(f90string(name, length));
	int id = Xmm::toDataMode(s);

	if (id < 0) {
		errstr<<"The string '"<<s<<"' is not among the list of known "
				"data modes: \n";
		for (unsigned n=0; n<=Xmm::N_DATA_MODES; ++n)
			errstr<<Xmm::toString(static_cast<Xmm::DataMode>(n))<<endl
				  <<Sas::error("Mismatch");
	}

	return id;
} // dataModeIdToInt


Integer3 CXXFUNC(filteridtoint, filterIdToInt)(const char *name,
											   Integer3 length)
{
	string s(f90string(name, length));
	int id = Xmm::toFilter(s);

	if (id < 0) {
		errstr<<"The string '"<<s<<"' is not among the list of known "
				"filter identifiers: \n";
		for (unsigned n=0; n<=Xmm::N_FILTERS; ++n)
			errstr<<Xmm::toString(static_cast<Xmm::Filter>(n))<<endl
				  <<Sas::error("Mismatch");
	}

	return id;
} // filterIdToInt


void CXXFUNC(instidtostring, instIdToString)
				(char *name, Integer3 length, const Integer3 *id)
{
	if (*id >= 0 && *id<=INT(Xmm::OM))
		stringf90(Xmm::toString((Xmm::Instrument)*id), name, length);
	else
		errstr<<"Instrument identifier outside allowed range [0.."
			  <<UNT(Xmm::OM)<<"]: "<<*id<<endl
			  <<Sas::error("OutOfBounds");
} // instIdToString


void CXXFUNC(modeidtostring, modeIdToString)
				(char *name, Integer3 length, const Integer3 *id)
{
	if (*id >= 0 && *id <= INT(Xmm::FLATFIELDING))
		stringf90(Xmm::toString((Xmm::Mode)*id), name, length);
	else
		errstr<<"Mode identifier outside allowed range [0.."
			  <<UNT(Xmm::FLATFIELDING)<<"]: "<<*id<<endl
			  <<Sas::error("OutOfBounds");
} // modeIdToString


void CXXFUNC(datamodeidtostring, dataModeIdToString)
				(char *name, Integer3 length, const Integer3 *id)
{
	if (*id >= 0 && *id <= INT(Xmm::SUMMARY_INFORMATION))
		stringf90(Xmm::toString((Xmm::DataMode)*id), name, length);
	else
		errstr<<"Data mode identifier outside allowed range [0.."
			  <<UNT(Xmm::SUMMARY_INFORMATION)<<"]: "<<*id<<endl
			  <<Sas::error("OutOfBounds");
} // dataModeIdToString


void CXXFUNC(filteridtostring, filterIdToString)
				(char *name, Integer3 length, const Integer3 *id)
{
	if (*id >= 0 && *id <= INT(Xmm::BARRED_U))
		stringf90(Xmm::toString((Xmm::Filter)*id), name, length);
	else if (*id < 0)
		stringf90("NONE", name, length);
	else
		errstr<<"Filter identifier outside allowed range [-1.."
			  <<UNT(Xmm::GRISM2)<<"]: "<<*id<<endl
			  <<Sas::error("OutOfBounds");
} // filterIdToString


Double CXXFUNC(stringtotimetag, stringToTimeTag)
				(const char *name, Integer3 length)
{
	return Xmm::toTimeTag(f90string(name, length));
} // stringtotimetag
