//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CrossPsf.h --- classes to model RGS PSF in Cross Dispersion Direction
//
//	DESCRIPTION:
//		Several classes are declared to describe the Cross Dispersion PSF
//		of the RGS
//
//	CHANGE HISTORY:
//		UL, 1999-09-16: 1.0.0: - first version
//		UL, 1999-11-25: 1.1.0: - changed I/F of CrossPsf::probability()
//
//	AUTHOR:
//		U. Lammers, Thu Sep 16 17:07:52 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Calibration Access and Data Handbook, C. Erd et. al
//		[3]: Calibration Access Layer, API guide, U. Lammers
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/CrossPsf.h,v 1.9 2003/10/03 14:15:41 sasbuild Exp $
//

#ifndef CROSS_PSF_H
#define CROSS_PSF_H

#include <vector>
#include <string>

#include "Dal.h"
#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalCommon.h"


//
//	---( CanonicalCrossPsf )--------------------------------------------------
//
class CanonicalCrossPsf : public CalAtom,
	public virtual CalChild<CalAtomCollection<CanonicalCrossPsf> > {
public:
	CanonicalCrossPsf(CalAtomCollection<CanonicalCrossPsf> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate a CanonicalCrossPsf object
		//


	static unsigned id() { return Ccf::CrossPSF; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual Spectrum probability(double time, double beta, double sourceXdisp,
								 const CalReal64Vector &xAngles,
								 double binWidth) const;
		//
		//	PURPOSE:
		//		Construct a normalized PSF probability distribution for the
		//		cross dispersion direction
		//
		//	ARGUMENTS:
		//		time		: time tag in TIMCOORD frame (see [3]) - currently
		//					  unused
		//		beta		: dispersion angle [rad]
		//		sourceXdisp	: position of celestial target in cross dispersion
		//					  direction [rad]
		//
		//	RETURNS:
		//		Normalized probability spectrum
		//


	virtual ~CanonicalCrossPsf();
		//
		//	PURPOSE:
		//		Destrory CrossPsf object constructed with above
		//		constructor(s).
		//
};


//
//	---( RgsCrossPsf )-------------------------------------------------------
//
class RgsCrossPsf : public virtual CanonicalCrossPsf {
public:
	RgsCrossPsf(const Table *crossPsfTable,
				unsigned channels,
				CalAtomCollection<CanonicalCrossPsf> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate RgsCrossPsf object from CCF table
		//

	void dump(ostream &os);
		//
		//	PURPOSE:
		//		see above
		//

	Spectrum probability(double time, double beta, double sourceXdisp,
						 const CalReal64Vector &xAngles,
						 double binWidth) const;
		//
		//	PURPOSE:
		//		see above
		//

	~RgsCrossPsf();
		//
		//	PURPOSE:
		//		see above
		//

private:
	std::vector<InterpolatedCurve> _gWidth, _gNorm, _gCenter,
								   _lWidth, _lNorm, _lCenter;

	double		_alphaIncidence, _gammaDesign, _rowlandRadius,
				_pixelSize1x1, _focalLength;

	unsigned	_channels1x1;
};

#endif

