//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	RgsBackgroundServer.cc --- implementation of RgsBackgroundServer class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 2004-7-10: 1.0.0: - initial version
//
//	AUTHOR:
//		R. Saxton
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/cal/src/RgsBackgroundServer.cc,v 1.1 2004/07/22 15:08:20 rsaxton Exp $
//
#include <iostream>

#include "RgsBackgroundServer.h"
#include "CalOalUtility.h"
#include "stringutil.h"

using namespace std;

//
//	---( RgsBackgroundServer> ) ---------------------------------------
//
RgsBackgroundServer::RgsBackgroundServer(DataSet *backset,
						CalAtomCollection<RgsBackgroundServer> *parent)
		: CalChild<CalAtomCollection<RgsBackgroundServer> >(parent)
{
    vector<unsigned> levelCount(3);

    unsigned nBlocks = backset->blocks();   

   // Loop over each table
    for (unsigned na=0; na<nBlocks; ++na) {

	    CcfConstituentTable ccfTabSpec(backset, na);
      const Table *table = ccfTabSpec.block();

      // Get the exposure, order and background level from attributes
        double exposure = table->attribute("EXPOSURE")->asReal();
        unsigned order = table->attribute("ORDER")->asInt();
        double level = table->attribute("LEVEL")->asReal();

      /* Read in the column data - coded longhand because the
         fillVector is not working */
        vector<int16> channel, quality;
        vector<int32> counts;
        vector<float> areascal, backscal;
              
      // Get the channel, counts. quality, areascal and backscal arrays
        //CalOalUtility::fillVector(table, "CHANNEL", channel);
        //CalOalUtility::fillVector(table, "COUNTS", counts);
        //CalOalUtility::fillVector(table, "QUALITY", quality);
        //CalOalUtility::fillVector(table, "AREASCAL", areascal);
        //CalOalUtility::fillVector(table, "BACKSCAL", backscal);


        ColumnData *colchan = table->column("CHANNEL")->data();
        ColumnData *colcounts = table->column("COUNTS")->data();
        ColumnData *colqual = table->column("QUALITY")->data();
        ColumnData *colarea = table->column("AREASCAL")->data();
        ColumnData *colback = table->column("BACKSCAL")->data();
        int16 *channel2 = colchan->int16Data();
        int32 *counts2 = colcounts->int32Data();
        int16 *quality2 = colqual->int16Data();
        real32 *area2 = colarea->real32Data();
        real32 *back2 = colback->real32Data();
        
        unsigned N = colchan->elements();
        for (unsigned n=0; n<N; ++n) {
            channel.push_back(channel2[n]);
            counts.push_back(counts2[n]);
            quality.push_back(quality2[n]);
            areascal.push_back(area2[n]);
            backscal.push_back(back2[n]);
        }

      // Remove the column data pointers
        delete colchan;
        delete colcounts;
        delete colqual;
        delete colarea;
        delete colback;
            
        _background.push_back( new RgsBackgnd(exposure, order, level,
                           channel, counts, quality, areascal, backscal) );

      // Keep a ruuning count of the number of levels available for each order
        _orderArray.push_back(order);
        _levelArray.push_back(levelCount[order]);
        levelCount[order]++;
    }
               
} // RgsBackgroundServer::RgsBackgroundServer()

void RgsBackgroundServer::dump(ostream &os)
{
	os<<"RgsBackgroundServer{\n";
	os<<"Channels\n"
	  <<"counts\n"<<' '<<endl;
    //vector<unsigned> chn = _background[0]->_channel();
    //vector<int> cnts = _background[0]->_counts();
	//for (unsigned n=0; n<chn.size(); n++)
		//os<<chn[n]<<' '<<cnts[n]<<' '<<endl;
	os<<"}";
} // RgsBackgroundServer::dump()


RgsBackgroundServer::~RgsBackgroundServer()
{
} // RgsBackgroundServer::~RgsBackgroundServer()


/* Method: index() - select the correct background */
const unsigned RgsBackgroundServer::index() const
{
    assert(_levelCalled);

    //cout<<"levReq "<<_levelRequired<<" ordReq "<<_orderRequired;
    //cout<<"orderArray size"<<_orderArray.size();

  // Find position in background array for the current order, index setting
    for (unsigned i=0; i<_orderArray.size(); i++) {
      if (_orderArray[i]==_orderRequired && _levelArray[i]==_levelRequired)
        {
         return i;   
        }
    }

    return 0;
    
} // RgsBackgroundServer::index

            
/* Method: numberExtensions() - get the number of extns for selected order */
const unsigned RgsBackgroundServer::numberExtensions() const
{
    if (_orderCalled)
    {
      unsigned count=0;
      
      for (unsigned i=0; i<_orderArray.size(); i++)
        if (_orderArray[i]==_orderRequired)
             count++;
        
      return count;
    }
    else {
      return ( _orderArray.size() ) ;
    }
                                                                                
} // RgsBackgroundServer::numberExtensions
                                                                                
/* Method: order(double) - set the grating order required */
const RgsBackgroundServer &RgsBackgroundServer::order(unsigned inputOrder)
{
    _orderCalled = true;
    _orderRequired = inputOrder;
                                                                                
    return *this;
} // RgsBackgroundServer::order

/* Method: level(unsigned) - set the background level required.
                           The level is input as an integer and is simply
                           an index into the array rather than the real level
*/
const RgsBackgroundServer &RgsBackgroundServer::level(unsigned inputLevel)
{
    if (inputLevel >= numberExtensions())
      errstr<<"Background level requested "<<inputLevel
            <<" outside range 0:"<<numberExtensions()-1
            <<Sas::error("BadBckgndLevel");

    _levelCalled = true;
    _levelRequired = inputLevel;

    return *this;
} // RgsBackgroundServer::level
                                                                                   
/* Method: exposure() - returns the exposure time for this backgnd map */
const double RgsBackgroundServer::exposure() const
{
    return _background[index()]->exposure();
} // RgsBackgroundServer::exposure()

/* Method: bckgndLevel() - returns the bckgnd level for this backgnd map */
const double RgsBackgroundServer::bckgndLevel() const
{
    return _background[index()]->level();
} // RgsBackgroundServer::bckgndLevel()

/* Method: channel() - returns the channel array for this backgnd map */
const vector<int16> RgsBackgroundServer::channel() const
{
    return _background[index()]->channel();
} // RgsBackgroundServer::channel()

/* Method: channel() - returns the counts array for this backgnd map */
const vector<int> RgsBackgroundServer::counts() const
{
    return _background[index()]->counts();
} // RgsBackgroundServer::counts()

/* Method: quality() - returns the quality array for this backgnd map */
const vector<int16> RgsBackgroundServer::quality() const
{
    return _background[index()]->quality();
} // RgsBackgroundServer::quality()

/* Method: areascal() - returns the areascal array for this backgnd map */
const vector<float> RgsBackgroundServer::areascal() const
{
    return _background[index()]->areascal();
} // RgsBackgroundServer::areascal()

/* Method: backscal() - returns the backscal array for this backgnd map */
const vector<float> RgsBackgroundServer::backscal() const
{
    return _background[index()]->backscal();
} // RgsBackgroundServer::backscal()




//
//	---( CalAtomCollection<RgsBackgroundServer> ) ----------------------
//
CalAtomCollection<RgsBackgroundServer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	RgsBackgroundServer *null = 0;
	_collection = new vector<RgsBackgroundServer *>(_nEntries, null);
	//
	//	Open the CCF file
	//
	CifEntry *e = ccf->entry(instrument, Ccf::TemplateBckgnd);
	DataSet *dataSet = e->open();

    // Instantiate the server object
	(*_collection)[0] = new RgsBackgroundServer(dataSet, this);


	dataSetServer->close(dataSet);
} // CalAtomCollection<RgsBackgroundServer>::CalAtomCollection()


RgsBackgroundServer *CalAtomCollection<RgsBackgroundServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<RgsBackgroundServer>::select


 //
//	---( RGS background class )---------------------------------------------------------------
//
RgsBackgnd::RgsBackgnd(double exposure, unsigned order, double level,
             vector<int16> channel, vector<int> counts,
             vector<int16> quality, vector<float>areascal,
             vector<float>backscal) :
                  _exposure(exposure), _order(order), _level(level),
                  _channel(channel), _counts(counts), _quality(quality),
                  _areascal(areascal), _backscal(backscal)
{
} // Constructor

/* Method: exposure() - returns the exposure time for this backgnd map */
const double RgsBackgnd::exposure() const
{
    return _exposure;
} // RgsBackgnd::exposure()

/* Method: level() - returns the background level for this backgnd map */
const double RgsBackgnd::level() const
{
    return _level;
} // RgsBackgnd::level()

 /* Method: channel() - returns the channel array for this backgnd map */
const vector<int16> RgsBackgnd::channel() const
{
    return _channel;
} // RgsBackgnd::channel()

   /* Method: counts() - returns the counts array for this backgnd map */
const vector<int> RgsBackgnd::counts() const
{
    return _counts;
} // RgsBackgnd::counts()

 /* Method: quality() - returns the quality array for this backgnd map */
const vector<int16> RgsBackgnd::quality() const
{
    return _quality;
} // RgsBackgnd::quality()

 /* Method: areascal() - returns the areascal array for this backgnd map */
const vector<float> RgsBackgnd::areascal() const
{
    return _areascal;
} // RgsBackgnd::areascal()

 /* Method: backscal() - returns the backscal array for this backgnd map */
const vector<float> RgsBackgnd::backscal() const
{
    return _backscal;
} // RgsBackgnd::backscal()

